/*  DDecon - Distributed Deconvolution Plugin for ImageJ
    Copyright (C) 2007  Ryan Littlefield

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

import ij.IJ;
import ij.ImagePlus;
import ij.Prefs;
import ij.gui.ImageCanvas;
import ij.gui.Plot;
import ij.gui.PlotWindow;
import ij.process.ImageProcessor;
import ij.util.Tools;

import java.awt.Color;
import java.awt.Point;
import java.awt.event.KeyEvent;
import java.awt.geom.Point2D;

public class DDplotter {

	public static final float[] ZERO = {0};
	public static double GRID_LINE_INTERVAL = Prefs.get("LSDDecon.gridIntvl",50);
	public static final double PLOT_W_PERPIXEL = Prefs.get("LSDDecon.pwpp",4.0);
	public static final double ERROR1 = 6.0;
	public static final double ERROR2 = 10.0;

	public DDwindow ddwin;
	public LSprocessor ls;
	public RepeatModel rm;
	Point2D.Double[] modelResidualPixels;
	Point2D.Double[] repeatResidualPixels;
	
	public ImagePlus modelProfileImp;
	public ImagePlus repeatProfileImp;
	public ImagePlus modelResidualImp;
	public ImagePlus repeatResidualImp;

	public double[] profileXbounds = new double [2];
	public double[] profileYbounds = new double [2];
	public int plotLEFT_MARGIN = 50;
	public int plotTOP_MARGIN = 20;
	public int plotHeight = 200;
	public int profilePlotWidth;
	public double profilePlotXscale;
	public double profilePlotYscale;
	public double reptScale;
	public double[] zeros2 = {0,0};
	public double[] modelXbounds = new double[2];
	public int modelPlotWidth;
	public double residualYbound;
	public double[][] modelScales;
	public Point2D.Double[] modelDistributionPoints;
	public int lineWidth;
	public LSDDecon_Manager manager;
//	public boolean autoNormalize = true;
	
	
	public DDplotter(LSprocessor ls, RepeatModel rm) {
		this(ls,rm,Prefs.get("LSDDecon.autoNorm",true));
	}
	
	public DDplotter(LSprocessor ls, RepeatModel rm, boolean autoNormalize) {
		this.ls = ls;
		this.rm = rm;

		if (autoNormalize) normalize();
		
//		calcRepeatResiduals();
//		calcModelResiduals();
//		calcPlotBounds();
		
		lineWidth = 1;
		modelProfileImp = new ImagePlus();
//		makeModelPlot(lineWidth);
		
		repeatProfileImp = new ImagePlus();
//		makeProfilePlot(lineWidth);
		
		repeatResidualImp = new ImagePlus();
//		makeRepeatResidualPlot(lineWidth);
		
		modelResidualImp = new ImagePlus();
//		makeModelResidualPlot(lineWidth);
		
		updatePlots(true);
	}
	
	public void normalize() {
		double scale = 100/getPerArea();
		for (int i=0; i<rm.repts.length; i++) {
			rm.repts[i].y *= scale;
		}
		rm.calcInts = rm.calcProfileInts(rm.modelProfilePixels,rm.repts);
	}
	
	public void updatePlots(boolean isModelChanged) {
		calcRepeatResiduals();
		calcModelResiduals();
		calcPlotBounds();
		if (isModelChanged) makeModelPlot(lineWidth);
		makeProfilePlot(lineWidth);
		makeRepeatResidualPlot(lineWidth);
		makeModelResidualPlot(lineWidth);
	}
	
	public void calcRepeatResiduals() {
		repeatResidualPixels = new Point2D.Double [rm.ddLength];
		for (int i=0; i<rm.ddLength; i++) {
			repeatResidualPixels[i] = new Point2D.Double(i,ls.lsInts[i]-rm.calcInts[i]);
		}
	}
		
	public void calcModelResiduals() {
		modelResidualPixels = new Point2D.Double[rm.ddLength];
		for (int i=0; i<rm.ddLength; i++) {
			modelResidualPixels[i] = new Point2D.Double(0,repeatResidualPixels[i].y);
		}
		int index = 0;
		double reptDLo,reptDHi,distLo,distHi; //,piqx,piqy;
		for (int i=0; i<=rm.repts.length; i++) {
			reptDLo = (i>0)?rm.repts[i-1].x:(-Double.MAX_VALUE);
			reptDHi = (i<rm.repts.length)?rm.repts[i].x:Double.MAX_VALUE;
			while (index<rm.ddLength && index>=reptDLo && index<reptDHi) {
				distLo = index - reptDLo;
				distHi = index - reptDHi;
				if (Math.abs(distLo)<Math.abs(distHi)) modelResidualPixels[index].x = distLo;
				else modelResidualPixels[index].x = distHi;
				index++;
			}
		}
	
	}
	
	public void calcPlotBounds() {
		// Repeat Profile X
		// Repeat Residual X
		profileXbounds[0] = -RepeatModel.MAX_OVERHANG;
		profileXbounds[1] = rm.ddLength+RepeatModel.MAX_OVERHANG;
		profilePlotWidth = (int)(PLOT_W_PERPIXEL*(rm.ddLength+2*RepeatModel.MAX_OVERHANG));
		profilePlotXscale = profilePlotWidth/(profileXbounds[1] - profileXbounds[0]);
		
		// Repeat Profile Y
		double[] lsBounds = Tools.getMinMax(ls.lsInts);
//		double[] rmBounds = Tools.getMinMax(rm.calcInts);
		profileYbounds[0] = 0;
//		profileYbounds[1] = Math.max(lsBounds[1],rmBounds[1]) * 1.1;
		profileYbounds[1] = lsBounds[1] * 1.1;
		profilePlotYscale = plotHeight/(profileYbounds[1] - profileYbounds[0]);
		reptScale = rm.psf.pixels[(rm.psf.pixels.length-1)/2].y;
		
		// Model Profile X
		// Model Residual X
		int n = rm.modelProfilePixels.length;
		modelXbounds[0] = rm.modelProfilePixels[0].x;
		modelXbounds[1] = rm.modelProfilePixels[n-1].x;
		modelPlotWidth = (int)(PLOT_W_PERPIXEL*n);
		
		// Model Residual Y
		// Repeat Residual Y
		double[] yMinMax = Tools.getMinMax(getYs(modelResidualPixels));
		residualYbound = Math.max(Math.abs(yMinMax[0]),Math.abs(yMinMax[1]));

	
		// Model Profile Y
		modelDistributionPoints = rm.model.makeDisplay();
		Point2D.Double[][] t = {rm.psf.pixels, modelDistributionPoints, rm.modelProfilePixels};
		double[][] toBeScaled = new double [t.length][];
		for (int i=0; i<t.length; i++) {
			toBeScaled[i] = new double [t[i].length];
			for (int j=0; j<t[i].length; j++) {
				toBeScaled[i][j] = t[i][j].y;
			}
		}
		modelScales = rescaleForPlot(toBeScaled, new Double(0), (Double)null);

	}
	
	public void makeModelPlot(int lineWidth) {
		Color lineColor = Color.blue;
		if (lineWidth>1) lineColor = Color.red;
		PlotWindow.plotWidth = modelPlotWidth;
		Plot modelProfilePlot = new Plot("modelProfilePlot","position","intensity",ZERO,ZERO);
		modelProfilePlot.setLineWidth(lineWidth);
		modelProfilePlot.setLimits(modelXbounds[0],modelXbounds[1],modelScales[3][0],modelScales[3][1]*1.2);
		modelProfilePlot.addPoints(zeros2,modelScales[3],Plot.LINE); // line for X-axis
		modelProfilePlot.setColor(Color.red);
		modelProfilePlot.addPoints(getXs(modelDistributionPoints),scale(getYs(modelDistributionPoints),modelScales[1]),Plot.CROSS);
		modelProfilePlot.setColor(lineColor);
		modelProfilePlot.addPoints(getXs(rm.modelProfilePixels),scale(getYs(rm.modelProfilePixels),modelScales[2]),Plot.LINE);
		modelProfilePlot.setColor(Color.black);
		modelProfilePlot.addPoints(getXs(rm.psf.pixels),scale(getYs(rm.psf.pixels),modelScales[0]),Plot.LINE);
		modelProfileImp.setProcessor(null,modelProfilePlot.getProcessor());
	}
	
	public void makeProfilePlot(int lineWidth) {
		Color lineColor = Color.blue;
		if (lineWidth>1) lineColor = Color.red;
		PlotWindow.plotWidth = profilePlotWidth;
		Plot repeatProfilePlot = new Plot("Repeat Profile Plot","position","intensity",ZERO,ZERO);
		repeatProfilePlot.setLineWidth(lineWidth);
		repeatProfilePlot.setLimits(profileXbounds[0],profileXbounds[1],profileYbounds[0],profileYbounds[1]);
		double[] reptsYscaled = new double [rm.repts.length];
		for (int i=0; i<rm.repts.length; i++) {
			reptsYscaled[i] = rm.repts[i].y*reptScale;
		}
		repeatProfilePlot.setColor(Color.red);
		repeatProfilePlot.addPoints(getXs(rm.repts),reptsYscaled,Plot.CROSS);
		repeatProfilePlot.setColor(lineColor);
		repeatProfilePlot.addPoints(rm.xvals,rm.calcInts,Plot.LINE);
		repeatProfilePlot.setColor(Color.black);
		repeatProfilePlot.addPoints(ls.xvals,ls.lsInts,Plot.LINE);
		repeatProfileImp.setProcessor(null,repeatProfilePlot.getProcessor());

	}
	
	public void makeRepeatResidualPlot(int lineWidth) {	
		PlotWindow.plotWidth = profilePlotWidth;
		Plot repeatResidualPlot = new Plot("Profile residuals","position","residual",ZERO,ZERO);
		repeatResidualPlot.setLineWidth(lineWidth);
		repeatResidualPlot.setLimits(profileXbounds[0],profileXbounds[1],-residualYbound,residualYbound);
		addGridLines(repeatResidualPlot,GRID_LINE_INTERVAL,profileXbounds,-residualYbound,residualYbound);
		repeatResidualPlot.setColor(Color.red);
		repeatResidualPlot.addPoints(getXs(repeatResidualPixels),getYs(repeatResidualPixels),Plot.CIRCLE);
		repeatResidualPlot.setColor(Color.black);
		repeatResidualImp.setProcessor(null,repeatResidualPlot.getProcessor());
	}
	
	public void makeModelResidualPlot(int lineWidth) {
		PlotWindow.plotWidth = modelPlotWidth;
		Plot modelResidualPlot = new Plot("TF residuals","position","residual",ZERO,ZERO);
		modelResidualPlot.setLineWidth(lineWidth);
		modelResidualPlot.setLimits(modelXbounds[0],modelXbounds[1],-residualYbound,residualYbound);
		addGridLines(modelResidualPlot,GRID_LINE_INTERVAL,modelXbounds,-residualYbound,residualYbound);
		modelResidualPlot.setColor(Color.red);
		modelResidualPlot.addPoints(getXs(modelResidualPixels),getYs(modelResidualPixels),Plot.CIRCLE);
		modelResidualPlot.setColor(Color.black);
		modelResidualImp.setProcessor(null,modelResidualPlot.getProcessor());
	}
	
	/** Adds horizontal grid lines to plot at a specified interval.  All grid lines are blue except for
	 *  one going through the origin which is black. */
	private void addGridLines (Plot plot, double interval, double[] xbounds, double miny, double maxy) {
		double yval = interval*(int)(miny/interval);
		plot.setColor(Color.black);
		while (yval<maxy) {
			plot.setColor((yval==0) ? Color.black : Color.blue);
			plot.addPoints(xbounds, new double [] {yval,yval}, Plot.LINE);
			yval+=interval;
		}
	}
	
	public double[] getXs(Point2D.Double[] pts) {
		double[] xs = new double [pts.length];
		for (int i=0; i<xs.length; i++) xs[i] = pts[i].x;
		return xs;
	}

	public double[] getYs(Point2D.Double[] pts) {
		double[] ys = new double [pts.length];
		for (int i=0; i<ys.length; i++) ys[i] = pts[i].y;
		return ys;
	}
	
	/** Linearly rescales a plot series.
	 * @param mb: an array of size two, containing slope and y-intercept values for linear scaling*/
	double[] scale(double[]x, double[] mb) {
		int n = x.length;
		double[] y = new double[n];
		for (int i=0; i<n; i++) y[i] = mb[0]*x[i] + mb[1];
		return y;
	}
	
	/**Rescales multiple series for plotting on the same graph.
	 * 
	 * @param series: an array of n series for plotting
	 * @param min: constant minimum value for all series, null value will autoscale 
	 * all series to their own minimum
	 * @param max: constant maximum value for all series, null value will autoscale
	 * all series to their own maximum
	 * @return a 2x(n+1) array of (m,b) values to linearly scale each series as well
	 * as the min and max values for the rescaled data in the last (m,b) pair.
	 */
	double[][] rescaleForPlot(double[][] series, Double min, Double max) {
		int n = series.length;
		if (n==0) return null;

		// Determine global min and max
		double [][] minmax = new double [n][2];
		minmax[0] = Tools.getMinMax(series[0]);
		double gmin = minmax[0][0];
		double gmax = minmax[0][1];
		for (int i=1; i<n; i++) {
			minmax[i] = Tools.getMinMax(series[i]);
			if (minmax[i][0]<gmin) gmin=minmax[i][0];
			if (minmax[i][1]>gmax) gmax=minmax[i][1];
		}

		// Determine m and b values
		double [][] mb = new double [n+1][2];

		//  No rescaling occurs and global min and max are returned 
		if ((n==1) || (min!=null && max!=null)) {
			for (int i=0; i<n; i++) {
				mb[i][0] = 1;
				mb[i][1] = 0;
			}
		}
		
		// Rescaled all series to go from gmin to gmax
		else if (min==null && max==null) { 
			for (int i=0; i<n; i++) {
				mb[i][0] = (gmax-gmin)/(minmax[i][1]-minmax[i][0]);
				mb[i][1] = gmin - mb[i][0]*minmax[i][0];
			}
		}
		
		// Rescale all series to go to gmax without changing min value
		else if (min!=null && max==null) {
			double set = min.doubleValue();
			gmin = gmax; // initialize rescaled gmin
			double lmin;
			for (int i=0; i<n; i++) {
				mb[i][0] = (gmax-set)/(minmax[i][1]-set);
				mb[i][1] = set*(1-mb[i][0]);
				lmin = minmax[i][0]*mb[i][0] + mb[i][1];
				if (lmin<gmin) gmin = lmin;
			}
		}
		
		// Rescale all series to go from gmin with out changing max value
		else if (min==null && max!=null) {
			double set = max.doubleValue();
			gmax = gmin; // initialize rescaled gmax
			double lmax;
			for (int i=0; i<n; i++) {
				mb[i][0] = (set-gmin)/(set-minmax[i][0]);
				mb[i][1] = set*(1-mb[i][1]);
				lmax = minmax[i][1]*mb[i][0] + mb[i][1];
				if (lmax>gmax) gmax = lmax;
			}
		}

		// Return global min and global max at end of mb array
		mb[n][0] = gmin;
		mb[n][1] = gmax;
		return mb;
	}
	
	/** Returns the percent area of the linescan accounted for by the model repeat profile. */
	public double getPerArea() {
		if (ls.lsInts == null || ls.lsInts.length == 0 || rm.calcInts == null) return Double.NaN;
		double realArea = 0;
		double calcArea = 0;
		for (int i=0; i<ls.lsInts.length; i++) {
			realArea += ls.lsInts[i];
			calcArea += rm.calcInts[i];
		}
		return (calcArea/realArea)*100;
	}
	
	/** Returns the correlation coefficient of the linescan and model repeat profile. */
	public double getCorrel() {
		if (ls.lsInts == null || ls.lsInts.length == 0 || rm.calcInts == null) return Double.NaN;
		int n = ls.lsInts.length;
		double sRC, sR, sR2, sC, sC2;
		sRC = sR = sR2 = sC = sC2 = 0;
		for (int i=0; i<n; i++) {
			sR += ls.lsInts[i];
			sR2 += ls.lsInts[i]*ls.lsInts[i];
			sC += rm.calcInts[i];
			sC2 += rm.calcInts[i]*rm.calcInts[i];
			sRC += ls.lsInts[i]*rm.calcInts[i];
		}
		double corr = (n*sRC-sR*sC)/Math.sqrt((n*sR2-sR*sR)*(n*sC2-sC*sC));
		return corr;
	}
	
	/** Returns the error between the linescan and the model repeat profile.  This error is 
	 *  essentially the sum of the square of the residuals, which is scaled according to the 
	 *  maximum intensity in the linescan and the number of data points. */
	public double getError() {
		if (ls.lsInts == null || ls.lsInts.length == 0 || repeatResidualPixels == null) return Double.NaN;
		int n = ls.lsInts.length;
		double sRes2 = 0;
		double maxInt = 0;
		for (int i=0; i<n; i++) {
			sRes2 += repeatResidualPixels[i].y*repeatResidualPixels[i].y;
			if (ls.lsInts[i]>maxInt) maxInt = ls.lsInts[i];
		}
		double error = 100*Math.sqrt(sRes2/n)/maxInt;
		return error;		
	}
	
	/** Returns a plot value (x,y) at a specified location on the canvas. */
	public Point realToPlot(Point reptReal) {
		Point rept = new Point();
		rept.x = (int)(profileXbounds[0] + (reptReal.x - plotLEFT_MARGIN)/profilePlotXscale);
		rept.y = (int)(profileYbounds[0] + (plotTOP_MARGIN + plotHeight - reptReal.y)/profilePlotYscale);
		rept.y /= reptScale;
		return rept;
	}
	
	/** Returns an array of locations (w,h) on the canvas corresponding to plot values (x,y).*/
	public Point[] plotToReal(Point[] repts) {
		Point[] reptReal = new Point [repts.length];
		int rx,ry;
		for (int i=0; i<repts.length; i++) {
			rx = 1 + plotLEFT_MARGIN + (int)((repts[i].x-profileXbounds[0])*profilePlotXscale);
			ry = plotTOP_MARGIN + plotHeight - (int)((repts[i].y*reptScale-profileYbounds[0])*profilePlotYscale);
			reptReal[i] = new Point(rx,ry);
		}
		return reptReal;
	}
	

}
