/*  DDecon - Distributed Deconvolution Plugin for ImageJ
    Copyright (C) 2007  Ryan Littlefield

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

import java.awt.*;
import java.awt.event.*;
import java.awt.geom.Point2D;
import java.io.*;
import java.util.*;
import java.awt.List;
import java.util.zip.*;
import ij.*;
import ij.process.*;
import ij.gui.*;
import ij.io.*;
import ij.text.*;
import ij.plugin.frame.*;
import ij.plugin.filter.*;

import ij.util.MacMenuBarSetter;
import ij.util.Tools;
import ij.macro.Interpreter;

/** This plugin implements the Analyze/Tools/ROI Manager command.
 * Called by:
 * 		DDeprocessor as declared field
 * 		DDeprocessor(LSinfo).<init>
 * 		DDeprocessor(LSinfo,SModel).<init>
 * 		DDwindow as declared field
 * 		LSDDecon_Manager.getInstance()
 * 		LSprocessor as declared field
 * 		LSprocessor(ImagePlus,LSinfo).<init>
 * 		LSWindow as declared field
 */
public class LSDDecon_Manager extends PlugInFrame implements ActionListener, ItemListener {

	public static final String VERSION = "3.01";
	public static final String DESCRIPTION = "ImageJ v" + IJ.getVersion() + ", LSDDecon v" + VERSION;
	static final int BUTTONS = 11;
	static final char FIXED='1', FREE='0';
	public static final String nameDelimiter = "_";
	static Frame instance;
	Panel linePanel, lsPanel, infoPanel;
	TextArea info;
	List list;
	public int lrIndex;
	Hashtable rois = new Hashtable();
	Line lineCopy;  // Copy of img.roi last added to list
//	Roi roiCopy;
	LSprocessor currLS;
	LSwindow lswin;
	RepeatModel currRM;
	DDplotter ddp;
	DDwindow ddwin;
	DDCurveFitter fitter;
	
	LSDDecon_Saver saver;
	LSDDecon_Loader loader;
	
	
	int slice2; // Image plane of last added img.roi
	TextField prefixField;
	Choice windowChoice;
	int[] imgID;
	TextWindow results = null; // DDeconResults window
	String lastUsedImageDir = "";
	String registerPath = IJ.getDirectory("startup");
	String subImageSavePath = IJ.getDirectory("startup");
	String intensitySavePath = IJ.getDirectory("startup");
	String resultsSavePath = "";//IJ.getDirectory("startup");
	CatalogedImageInfo catImageInfo;
	String subImageInfo = null;
//	LSwindow LSwin;
//	DDwindow DDwin;
	Catalog_ catalog;
	boolean canceled;
	boolean macro;
	boolean ignoreInterrupts;

	/** Initiates a new LSDDecon_Manager if one is not open, brings it to the front if it is already open.
	 * Called by: NONE
	 *
	 * The plugin menu command instantiates this object.
	 */
	public LSDDecon_Manager() {
		super("LSDDecon Manager");
		if (instance!=null) {
			instance.toFront();
			return;
		}
		instance = this;
		ImageJ ij = IJ.getInstance();
 		addKeyListener(ij);
		WindowManager.addWindow(this);
		IJ.setTool(4); // Selects straight line tool
		setLayout(new FlowLayout(FlowLayout.CENTER,5,5));
		
		// Line panel, containing the list of lines and LSinfo objects
		linePanel = makeLinePanel(ij);
		add(linePanel);

		// Linescan panel, containing textfield for prefix ID, stored info display region, and image list
		lsPanel = new Panel();
		lsPanel.setLayout(new BorderLayout());
		lsPanel.add(makePrefixPanel(),BorderLayout.NORTH);
		makeInfoPanel();
		windowChoice = new Choice();
		windowChoice.addItemListener(this);
		updateWindowList();
		lsPanel.add(windowChoice,BorderLayout.SOUTH);
		add(lsPanel);
		
		pack();
		list.remove(0); // remove the width-setter
		Rectangle r = ij.getBounds();
		setLocation(r.x+r.width-getWidth(),r.y+r.height);
		show();
	}
	
	/** Makes the panel for listing and manipulating stored lines and LSinfo objects
	 * Called by:
	 * 		LSDDecon_Manager().<init>
	 * 
	 * @param ij: this imageJ, as a KeyListener.
	 * @return panel with list and buttons
	 */
	Panel makeLinePanel(ImageJ ij) {
		int rows = 15;
		boolean allowMultipleSelections = IJ.isMacintosh();
		list = new List(rows, allowMultipleSelections);
		list.add("01234567890123456789012345678901"); // This ultimately sets the width of the list
		list.addItemListener(this);
 		list.addKeyListener(ij);
 		list.setSize(200,300);
		add(list);
		Panel p = new Panel();
		int nButtons = BUTTONS;
		p.setLayout(new GridLayout(nButtons, 1, 5, 0));
		p.add(aButton("Catalog Image"));
		p.add(aButton("Add Line"));
		p.add(aButton("Apply Selection"));
		p.add(aButton("Do LineScan"));
		p.add(aButton("Do DDecon"));
		p.add(aButton("Do Fitting"));
		p.add(aButton("Rename Selection"));
//		p.add(aButton("Open LineScan"));
		p.add(aButton("Save"));
		p.add(aButton("Delete Selection"));
		p.add(aButton("Open Results"));
		p.add(aButton("Load LineScan"));
//		p.add(aButton("Load Result"));
//		p.add(aButton("Deselect"));
//		p.add(aButton("Export Data"));
		return p;
	}
	
	/** Makes the panel for specifying prefix for linescans.
	 * Called by:
	 * 		LSDDecon_Manager().<init>
	 * 
	 * Set-up to get prefix from Prefs.
	 * 
	 * @return panel with Label and TextField
	 */
	Panel makePrefixPanel() {
		Panel p = new Panel();
		Label labPrefix = new Label("ID prefix:");
		p.add(labPrefix);
		prefixField = new TextField();
		prefixField.setText(Prefs.get("LSDDecon.lsprefix","LS"));
		p.add(prefixField);		
		return p;
	}

	/** Makes panel with line or LSinfo information.
	 * Called by:
	 * 		LSDDecon_Manager().<init>
	 *
	 */
	void makeInfoPanel() {
		if (infoPanel != null) {
			lsPanel.remove(infoPanel);
		}
		infoPanel = new Panel();
		infoPanel.setLayout(new FlowLayout(FlowLayout.CENTER,5,5));
		
		String text = "";
		info = new TextArea(text,10,15,TextArea.SCROLLBARS_NONE);
//		info = new TextArea(text,10,20,TextArea.SCROLLBARS_BOTH);
		info.setEditable(false);
		infoPanel.add(info);
		lsPanel.add(infoPanel,BorderLayout.CENTER);
		lsPanel.repaint();
	}
	
	/** Active, labeled button.
	 * Called by:
	 * 		LSDDecon_Manager.makeLinePanel(ImageJ)
	 * ActionListener set to this LSDDecon_Manager.
	 * KeyListener set to IJ.getInstance().
	 * 
	 * @param label: button label
	 * @return active button
	 */
	Button aButton(String label) {
		Button b = new Button(label);
		b.addActionListener(this);
		b.addKeyListener(IJ.getInstance());
		return b;
	}

	public void actionPerformed(ActionEvent e) {
//		System.out.println(e.paramString());
		int modifiers = e.getModifiers();
		boolean altKeyDown = (modifiers&ActionEvent.ALT_MASK)!=0 || IJ.altKeyDown();
		boolean shiftKeyDown = (modifiers&ActionEvent.SHIFT_MASK)!=0 || IJ.shiftKeyDown();
		IJ.setKeyUp(KeyEvent.VK_ALT);
		IJ.setKeyUp(KeyEvent.VK_SHIFT);
		String label = e.getActionCommand();
		if (label==null)
			return;
		String command = label;
		if (command.equals("Catalog Image"))
			catalogImage2(getImage());
		else if (command.equals("Add Line"))
			add(false);
//			add(shiftKeyDown, altKeyDown);
		else if (command.equals("Apply Selection"))
			applyToImage();
		else if (command.equals("Do LineScan"))
			doLineScan();
		else if (command.equals("Do DDecon"))
			doDDecon();
		else if (command.equals("Do Fitting"))
			doFitting();
		else if (command.equals("Rename Selection"))
			rename();
		else if (command.equals("Open LineScan"))
			open(null);
		else if (command.equals("Save"))
			save();
		else if (command.equals("Delete Selection"))
			delete(false);
		else if (command.equals("Open Results"))
			openResults2("");
		else if (command.equals("Load LineScan"))
			loadLineScan();
		else if (command.equals("Load Result"))
			loadResultsLine2();
		else if (command.equals("Deselect"))
			deselect();
		else if (command.equals("Export Data"))
			exportData();
	}

	public void itemStateChanged(ItemEvent e) {
//		System.out.println(e.paramString());
		if (e.getSource() == list) {
			if (e.getStateChange()==ItemEvent.SELECTED 
//			&& WindowManager.getCurrentImage()!=null && !ignoreInterrupts
			) {
				int index = 0;
	            try {index = Integer.parseInt(e.getItem().toString());}
	            catch (NumberFormatException ex) {}
				if (index<0) index = 0;
//				if (index>=0) {
					String label = list.getItem(index);
					String text = getInfoText(label);
					info.setText(text);
//				}
			}
		}
		else if (e.getSource() == windowChoice) { // XXX May not need this if getImage is used
			if (e.getStateChange()==ItemEvent.SELECTED) {
				int index = windowChoice.getSelectedIndex();
				int id = this.imgID[index];
				ImagePlus imp = WindowManager.getImage(id);
			}
		}
	}
	
	// XXX from pluginframe, want to listen to all open linescan and ddecon windows
    public void windowActivated(WindowEvent e) {
		if (IJ.isMacintosh() && IJ.getInstance()!=null)
			new MacMenuBarSetter((Frame)this);
		WindowManager.setWindow(this);
		
		// Update choice of windows
		updateWindowList();
	}
    
    public void updateWindowList() {
		windowChoice.removeAll();
		imgID = WindowManager.getIDList();
		if (imgID!=null) {
			for (int i=0; i<imgID.length; i++) {
				windowChoice.add(WindowManager.getImage(imgID[i]).getTitle());
			}
		}
		ImagePlus imp = WindowManager.getCurrentImage();
		if (imp != null) windowChoice.select(imp.getTitle());
    }
    
    public void windowClosed(WindowEvent e) {
    	Window w = e.getWindow();
    	if (w instanceof TextWindow && ((TextWindow)w).getTitle().equals(results.getTitle())) {
    		results = null;
    	}
//    	System.out.println("Manager.windowClosed: results="+results);
    }
    
    /** Enters a copy of an LSinfo object into the hashtable of lines
     * Called by:
     * 		LSprocessor.updatePlots(ImageProcessor)
     * 
     * This will overwrite a previously stored LSinfo or line with the same linescanID.
     * 
     * @param lsi: LSinfo object to enter into hashtable
     */
/*	public void setLSinfo(LSinfo lsi) { // XXX Need to work on this method
		String label = lsi.linescanID;
		if (rois.get(label)!=null) {
			rois.remove(label); // XXX Update roi in hashtable with the same name (not unique)
		}
		else {
			list.add(label);
		}
		LSinfo lsiCopy = (LSinfo)lsi.clone();
		lsiCopy.linescanID = label;
		rois.put(label, lsiCopy);
	}
*/	
	/** Gets an LSinfo object from the list of lines and LSinfo objects stored in the list.
	 * Called by:
	 * 		LSDDecon_Manager.doLineScan()
	 * 
	 * Line (ROI) information is used and combined with an imagename to start LSprocessor
	 *
	 * @param label: String stored in rois hashtable.
	 * @return LSinfo object generated from the stored line of LSinfo object.
	 */
/*	public LSinfo makeLSInfo(String label) {
		ImagePlus imp = getImage(); // Gets the selected or current image
		if (imp==null) return null;
		
		LSinfo lsi;
		// The line list is used to generate a raw (ROI info only) or saved LSinfo object 
		Object o = rois.get(label);
		if (o instanceof Line) {
			// A new roi is made into an LSinfo object and passed
			Line line = (Line)o;
			lsi = new LSinfo(WindowManager.getImage(imgID[windowChoice.getSelectedIndex()]),line);

			// LinescanID is not provided for stored lines, so one is generated.
			String lsIDprefix = this.prefixField.getText();
			if (lsi.linescanID == null) lsi.linescanID = getUniqueLSName(lsIDprefix+"-1");
		}
		else {
			// A saved LSinfo object is passed with all of its presets
			lsi = (LSinfo)o;
			lsi.fi = getImage().getFileInfo();
		}

		// The selected image name is combined with the LSinfo object
//		ImagePlus imp = getImage();
//		lsi.fi.fileName = imp.getShortTitle();

		// LinescanID is not provided for stored lines, so one is generated.
//		String lsIDprefix = this.prefixField.getText();
//		if (lsi.linescanID == null) lsi.linescanID = getUniqueLSName(lsIDprefix+"-1");
		return lsi;
	}
*/	
	/** Gets a description of the line or LSinfo object stored in the hashtable.
	 * Called by:
	 * 	LSDDecon_Manager.add(boolean)
	 * 	LSDDecon_Manager.itemStateChanged(ItemEvent)
	 * 	LSDDecon_Manager.loadResultsLine()
	 * 
	 * @param label: String stored in rois hashtable
	 * @return Formated string of values describing the object.
	 */
	public String getInfoText(String label) {
		String text;
		Object o = rois.get(label);
//		Line line;
//		line = (Line)o;
		LineRegion lr = (LineRegion)o;
		Line line = lr.line;
		text = "Line at"
  			+ "\np1 = (" + IJ.d2s(line.x1d,2)
  			+ "," + IJ.d2s(line.y1d,2)
  			+ ")\np2 = (" + IJ.d2s(line.x2d,2)
  			+ "," + IJ.d2s(line.y2d,2)
  			+ ")\nR0 width = " + IJ.d2s(lr.regWidths[0],0)
  			+ "\nR0 mode = " + LSRegion.getModeString(lr.regModes[0])
  			+ "\nR1 width = " + IJ.d2s(lr.regWidths[1],0)
  			+ "\nR1 mode = " + LSRegion.getModeString(lr.regModes[1])
  			+ "\nR2 width = " + IJ.d2s(lr.regWidths[2],0)
  			+ "\nR2 mode = " + LSRegion.getModeString(lr.regModes[2])
  			+ "\nSkewAngle = " + IJ.d2s(lr.skewAngle,2);
  			;
		return text;
	}
	
	public void catalogImage2(ImagePlus imp) {
		if (imp==null) return;
		saver = new LSDDecon_Saver(this);
		saver.catalogImage(imp);
	}

/*	public void catalogImage(ImagePlus imp) {
//		ImagePlus imp = getImage();
		if (imp==null) return;
		// Check to see if this image was already cataloged by looking for text file in the same directory.
		String dir = imp.getOriginalFileInfo().directory;
		String fname = imp.getShortTitle();
		String path = dir + fname + CatalogedImageInfo.POSTFIX;
		File f = new File(path);
		catImageInfo = new CatalogedImageInfo(dir,fname,f.exists());
	}
*/	
	/** Instantiates a new LSprocessor using the selected or active image and selected line in the list.
	 * Called by:
	 * 		LSDDecon_Manager.actionPerformed(ActionEvent)
	 * 
	 */
	public boolean doLineScan() {
		lrIndex = list.getSelectedIndex();
		if (lrIndex<0)
			return error("Exactly one item in the list must be selected.");
		String name = list.getItem(lrIndex);
		Object o = rois.get(name); // Gets the selected line
		ImagePlus imp = getImage(); // Gets the selected or current image
		if (imp==null) { // No image: send error and return false
			return error("Cannot create a LineScan from a line without an open image");
		}
		else { // An image is selected: make a new line for linescan
			LineRegion lr = (LineRegion)o;
			currLS = new LSprocessor (imp,lr);
//			Line line = (Line)((Line)o).clone();
//			currLS = new LSprocessor (imp,line);
			currLS.manager = this;
			return true;
		}	
				
	}
	
	public boolean doDDecon() {
		if (currLS != null) {
			currRM = new RepeatModel(currLS.lsLength);
			ddp = new DDplotter(currLS,currRM);
			ddp.manager = this;
			ddwin = new DDwindow(ddp);
		}
		return true;
	}
	
	public boolean doFitting() {
		IJ.showStatus("Fitting Profile....");
		
//		currRM.setParamStates();
		boolean[][] pStates = new boolean [3][];
		pStates[0] = ddwin.getModelParamStates();
//		pStates[0] = currRM.model.getParamStates();
//		for (int i=0; i<pStates[0].length; i++) {
//			System.out.println("Manager.doFitting: pStates[0]["+i+"]="+pStates[0][i]);
//		}
		pStates[1] = new boolean [currRM.repts.length];
		pStates[2] = new boolean [currRM.repts.length];
		for (int i=0; i<currRM.repts.length; i++) {
			pStates[1][i] = currRM.cbReptPositions[i].getState();
			pStates[2][i] = currRM.cbReptWeights[i].getState();
//			System.out.println("Manager.doFitting: pStates[1]["+i+"]="+pStates[1][i]);
		}

		fitter = new DDCurveFitter(currRM, currLS.lsInts);
//		fitter.pStates[0] = pStates[0];
//		fitter.pStates[1] = pStates[1];
//		fitter.pStates[2] = pStates[2];
		boolean fit = fitter.doFit(true);
		if (fit) {
			IJ.showStatus("Fit Completed");
			IJ.beep();

			// Update the plots and display the results of the fit
			double[] nparams = fitter.decodeParams(fitter.getParams());
			
			int nmp = currRM.model.getParamNumber();
			double[] modelParams = new double [nmp];
			for (int i=0; i<nmp; i++) {
				modelParams[i] = nparams[i];
			}
			currRM.model.setParams(modelParams);
//			currRM.psf.focusParam = modelParams[0];

			int nr = currRM.repts.length;
			Point2D.Double[] repts = new Point2D.Double [nr];
			for (int i=0; i<nr; i++) {
				repts[i] = new Point2D.Double(nparams[nmp+i],nparams[nmp+nr+i]);
			}
			currRM.repts = repts;
			
			currRM.modelProfilePixels = currRM.calcModelInts(currRM.model.getParams());
			currRM.calcInts = currRM.calcProfileInts(currRM.modelProfilePixels,currRM.repts);

			ddp.updatePlots(true);

			for (int i=0; i<nmp; i++) {
				ddwin.tfModelParams[i].setText(IJ.d2s(modelParams[i],2));
			}
			for (int i=0; i<nr; i++) {
				currRM.tfReptPositions[i].setText(IJ.d2s(repts[i].x,2));
				currRM.tfReptWeights[i].setText(IJ.d2s(repts[i].y,2));
			}

			ddwin.updateWindow();
		    
//			fitter.setInfo(this);
	    	String fitMethod = DDCurveFitter.fitList[0];//"Simplex";
	    	int maxIter = fitter.maxIter;
	    	int restarts = fitter.setRestarts;
	    	double errTol = fitter.maxError;
	    	String infoFit = fitter.info;
	    	double max_int_search = fitter.maxIntSearch;
	    	int max_overhang = RepeatModel.MAX_OVERHANG;
	    	double max_pos_search = fitter.maxPosSearch;
	    	int numIter = fitter.numIter;
	    	double goodness = fitter.getFitGoodness();
	    	double ssr = fitter.getSumResidualsSqr();
	    	double chi2 = fitter.getChiSquared();

			
			
			
			
			Prefs.set("LSDDecon.gridIntvl",DDplotter.GRID_LINE_INTERVAL); // XXX not user adjustable yet
			Prefs.set("LSDDecon.autoNorm",true); // XXX Not user adjustable yet
			Prefs.set("LSDDecon.pwpp",DDplotter.PLOT_W_PERPIXEL); // XXX Not user adjustable yet
			Prefs.set("LSDDecon.minTFL",fitter.minTFL);
			Prefs.set("LSDDecon.maxTFL",fitter.maxTFL);
			
			Prefs.set("LSDDecon.maxOverhang",RepeatModel.MAX_OVERHANG); // XXX not user adjustable yet
			//Prefs.set("LSDDecon.defModIndex",ddp.rm.cModIndex); // XXX Set to current model used
			Prefs.set("LSDDecon.pSearchR",fitter.max_pos_search);
			Prefs.set("LSDDecon.wSearchR",fitter.max_int_search);
			double aveSL = RepeatModel.REPEAT_SPACING;
//			int nr = ddp.rm.repts.length;
			if (nr > 1)
				aveSL = (ddp.rm.repts[nr-1].x - ddp.rm.repts[0].x)/(nr-1); // average sarcomere length
			Prefs.set("LSDDecon.aveSL", aveSL);
			Prefs.set("LSDDecon.linescanID",ddwin.lfPFID.getText());
			Prefs.set("LSDDecon.date",ddwin.lfDate.getText());
			Prefs.set("LSDDecon.sample",ddwin.lfSample.getText());
			Prefs.set("LSDDecon.fluor",ddwin.lfFluor.getText());
			Prefs.set("LSDDecon.exposure",Double.parseDouble(ddwin.lfExpT.getText()));
//			ddi.savePreferences();
		}
		return true;
	}
	
	public void exportData() {
		IJ.showStatus("Saving Data...");
		saver = new LSDDecon_Saver(this);
		saver.exportAll();
		updateLR();
//		saver.exportData();

/*		boolean[] saves = confirmExport();
		if(saves[0]) {
			// Save Cataloged Image
			IJ.showStatus("Saving Cataloged Image...");
			catImageInfo.save();
		}
		if(saves[1]) {
			// Save SubImage
			IJ.showStatus("Saving SubImage...");
			saveSubImage();
		}
		if (saves[2]) {
			// Save Intensities in a text file
			IJ.showStatus("Saving Intensities...");
			saveIntensities(intensitySavePath);
			doSarcView();
		}
		if (saves[3]) {
			// Save Results
			IJ.showStatus("Saving Results...");
			doExport();
//			results.getTextPanel().saveAs(resultsSavePath);
		}
*/
	}
	
	public void updateLR() {
		String name = list.getItem(lrIndex);
		String name2 = name;
//		String name2 = currLS.lsName;
		Object o = rois.get(name);
		LineRegion lr = (LineRegion)o;
		for (int i=0; i<currLS.regions.length; i++) {
			lr.regWidths[i] = currLS.regions[i].rect.height;
			lr.regModes[i] = currLS.regions[i].mode;
		}
		lr.bgCorrIndex = currLS.bgCorrIndex;
		lr.skewAngle = currLS.skewAngle;
		rois.remove(name);
		lr.line.setName(name2);
		rois.put(name2, lr);
		list.replaceItem(name2, lrIndex);
		list.select(lrIndex);
//		System.out.println("manager.updateLR: lr.line= "+lr.line);
		String text = getInfoText(name2);
		info.setText(text);

	}
	
	public void updateLR(Line l) {
		String name = list.getItem(lrIndex);
		Object o = rois.get(name);
		LineRegion lr = (LineRegion)o;
		rois.remove(name);
		lr.line = l;
		lr.line.setName(name);
		rois.put(name, lr);
		list.replaceItem(name, lrIndex);
		list.select(lrIndex);
//		System.out.println("manager.updateLR: lr.line= "+lr.line);
		String text = getInfoText(name);
		info.setText(text);

	}
	
    public void openResults2(String path) {
    	loader = new LSDDecon_Loader(this);
    	loader.openResults(path);
    }
    
    public void loadLineScan() {
    	loader = new LSDDecon_Loader(this);
    	loader.loadLineScan();
//    	currLS = loader.loadLineScan();
//    	add(false);
    }

	public void loadResultsLine2() {
		loader = new LSDDecon_Loader(this);
		loader.loadResultsLine(0);
	}
	
/*	public boolean[] confirmExport() {
		// Open confirmation dialog
		GenericDialog gd = new GenericDialog("Confirm Data Export");
		gd.addMessage("Original Image");
		gd.addMessage(getCatalogInfo());
		gd.addCheckbox("Catalog Image",false);
		gd.addStringField("Location: ",registerPath,90);
		
		gd.addMessage("LineScan SubImage");
		gd.addMessage(prepareSubImage());
		gd.addCheckbox("Save",true);
		gd.addStringField("Location: ",subImageSavePath,90);
		
		gd.addMessage("LineScan, Modeled, and SarcView Intensities");
		gd.addMessage(prepareIntensities());
		gd.addCheckbox("Save",false);
		gd.addStringField("Location: ",intensitySavePath,90);
		
		gd.addMessage("Results");
		gd.addMessage(prepareResults());
		gd.addCheckbox("Save",false);
		gd.addStringField("Location: ",resultsSavePath,90);
		gd.showDialog();

		boolean[] saves = {false,false,false,false};
		if (gd.wasCanceled()) return saves;
		saves[0] = gd.getNextBoolean();
		registerPath = gd.getNextString();
		saves[1] = gd.getNextBoolean();
		subImageSavePath = gd.getNextString();
		saves[2] = gd.getNextBoolean();
		intensitySavePath = gd.getNextString();
		saves[3] = gd.getNextBoolean();
		resultsSavePath = gd.getNextString();
		Prefs.set("LSDDecon.lastResultsPath",resultsSavePath);
		return saves;
	}
*/	
/*	public String getCatalogInfo() {
		if (catImageInfo==null && getImage()!=null) catalogImage(getImage());
		String s = "" + this.catImageInfo.confirm();
		registerPath = this.catImageInfo.getSavePath();
		return s;
	}
*/	
/*	public String prepareSubImage() {
		if (subImageInfo==null) setSubImageInfo();
		FileInfo fi = currLS.subImage.getOriginalFileInfo();
		subImageInfo = "Name= " + fi.fileName
				+ "\nWidth = " + fi.width + " pixels (" + IJ.d2s(fi.width*fi.pixelWidth,2) + " microns)"
				+ "\nHeight = " + fi.height + " pixels (" + IJ.d2s(fi.height*fi.pixelHeight,2) + " microns)"
				+ "\nInfo = " + fi.info
				+ "\nDescription = " + fi.description
				;
		this.subImageSavePath = fi.directory + fi.fileName;
		return subImageInfo;
	}
*/	
/*	public String prepareIntensities() {
		String intensityInfo;
		if (ddwin==null) {
			// No DDecon, so only linescan intensities will be written
			intensityInfo = "Only Linescan data avaliable.";
		}
		else {
//			DDinfo ddi = this.DDwin.ddi;
			// Observed Linescan Intensities
			double[] Iobs = currLS.lsInts;
			int[] intCalc = {currLS.regions[0].mode, currLS.regions[1].mode, currLS.regions[2].mode};
			int corrCalc = currLS.bgCorrIndex;
			
			// Modeled DDecon Intensities
			int nr = currRM.repts.length;
			Point2D.Double[] repts = currRM.repts;
			intensityInfo = "Number of points=" + Iobs.length
					+ "\nBG1calc=" + intCalc[0]
					+ "; LScalc=" + intCalc[1]
					+ "; BG2calc=" + intCalc[2]
					+ "\nBGcorrection=" + corrCalc
					+ "\nNumber of Repeats=" + nr
					+ "\nAverage Sarcomere Length=" + IJ.d2s((repts[nr-1].x-repts[0].x)/(nr-1),2)+" pixels.";
					;
			intensitySavePath = currLS.imp.getOriginalFileInfo().directory +
								currLS.subImage.getShortTitle() +
								"_int.txt";
		}
		return intensityInfo;
	}
*/	
/*	public String prepareResults() {
//		DDinfo ddi = this.DDwin.ddi;
		String s = "";
		resultsSavePath = currLS.imp.getOriginalFileInfo().directory + "Distributed Deconvolution Results.txt";
		return s;
	}
*/	
/*    public void saveSubImage() {
    	setSubImageInfo();
    	FileSaver fs = new FileSaver(currLS.subImage);
    	fs.save();
    }
*/    
/*    public void setSubImageInfo() {
    	FileInfo fi = currLS.subImage.getFileInfo();
    	fi.directory = currLS.imp.getOriginalFileInfo().directory;
    	fi.fileName = currLS.subImage.getTitle();
    	fi.fileFormat = FileInfo.TIFF;
    	fi.description = LSDDecon_Manager.DESCRIPTION;
    	fi.info = "skewAngle=" + currLS.skewAngle
    			+ "; regionWidths=[" + currLS.regions[0].rect.height
    			+ "," + currLS.regions[1].rect.height
    			+ "," + currLS.regions[2].rect.height
    			+ "]"
    			+ "; line: x1d,y1d=(" + IJ.d2s(currLS.line.x1d,2)
    			+ "," + IJ.d2s(currLS.line.y1d,2)
    			+ "); x2d,y2d=(" + IJ.d2s(currLS.line.x2d,2)
    			+ "," + IJ.d2s(currLS.line.y2d,2)
    			+ ")";
    	currLS.subImage.setFileInfo(fi);
    }
*/    
/*	public void saveIntensities(String path) {
    	String name = "Intensities";
    	String intHeadings = "Position\tI_Obs\tI_Calc\tCalcData";
    	String intValues;
    	TextWindow win = new TextWindow(name,"",300,200);
    	TextPanel tp = win.getTextPanel();
   		tp.setColumnHeadings(" " + intHeadings);

    	for (int i=0; i<currLS.lsInts.length; i++) {
       		intValues = i + "\t" + IJ.d2s(currLS.lsInts[i],4) + "\t" + IJ.d2s(currRM.calcInts[i],4) + "\t";
    		if (i==0) intValues = intValues + currLS.regions[0].mode + "\t";
    		if (i==1) intValues = intValues + currLS.regions[1].mode + "\t";
    		if (i==2) intValues = intValues + currLS.regions[2].mode + "\t";
    		if (i==3) intValues = intValues + currLS.bgCorrIndex + "\t";
    		tp.appendLine(intValues);
    	}
    	tp.saveAs(path);
    	win.close();
	}
*/	
/*	public boolean doSarcView() {
		IJ.log("Starting sarcView");
//		ddi.pStates = getParamStates(); // set the current param fix states in ddecon processor
//		if (ddi==null) ddi = new DDinfo();
		SarcData[] sd = prepSarcView();
		
		// Make new ResultWindow with the view results
		double p;
		double q;
		TextWindow tw = new TextWindow("","",100,100);
		WindowManager.addWindow(tw);
		tw.setVisible(false);
		TextPanel tp = tw.getTextPanel();
		String path;
		String headings = "SarcData\tDistance From Z\tDistance From M\tIobs\tIcalc\tIres";
		tp.setColumnHeadings(headings);
		String lineOut;
		for (int n=0; n<sd.length; n++) {
			for (int i=0; i<sd[n].npix; i++) {
				lineOut = " " + "\t";
				p = sd[n].length/2 - sd[n].xVals[i];
				q = sd[n].iVals[0][i]-sd[n].iVals[1][i];
				if (i==0) lineOut = IJ.d2s(n,0) + "\t";
				if (i==1) lineOut = IJ.d2s(sd[n].length,4) + "\t";
				if (i==2) lineOut = IJ.d2s(currRM.repts[n].y,4) + "\t";
				if (i==3) lineOut = IJ.d2s(currRM.repts[n+1].y,4) + "\t";
				lineOut = lineOut + IJ.d2s(sd[n].xVals[i],4) + "\t"
								+ IJ.d2s(p,4) + "\t" 
								+ IJ.d2s(sd[n].iVals[0][i],4) + "\t" 
								+ IJ.d2s(sd[n].iVals[1][i],4) + "\t"
								+ IJ.d2s(q,4) + "\t";
				tp.appendLine(lineOut);
			}
			String title = "Sarcomere-"+n;
			tw.setTitle(title);
			tw.show();
//			results.show(title);
			path = currLS.imp.getOriginalFileInfo().directory + currLS.subImage.getShortTitle() + "_sarc"+n+".txt";
			IJ.log("DDwindow.doSarcView: path= "+path);
//			path="";
			PrintWriter pw = null;
			try {
				FileOutputStream fos = new FileOutputStream(path);
				BufferedOutputStream bos = new BufferedOutputStream(fos);
				pw = new PrintWriter(bos);
			}
			catch (IOException e) {
				IJ.write("" + e);
			}
			tp.save(pw);
			pw.close();
			tp.selectAll();
			tp.clearSelection();
			tp.resetSelection();
		}
		tw.close();

		
		Prefs.set("LSDDecon.gridIntvl",DDplotter.GRID_LINE_INTERVAL); // XXX not user adjustable yet
		Prefs.set("LSDDecon.autoNorm",true); // XXX Not user adjustable yet
		Prefs.set("LSDDecon.pwpp",DDplotter.PLOT_W_PERPIXEL); // XXX Not user adjustable yet
		Prefs.set("LSDDecon.minTFL",fitter.minTFL);
		Prefs.set("LSDDecon.maxTFL",fitter.maxTFL);

		Prefs.set("LSDDecon.maxOverhang",repeatModel.MAX_OVERHANG); // XXX not user adjustable yet
		Prefs.set("LSDDecon.defModIndex",currRM.cModIndex); // XXX Set to current model used
		Prefs.set("LSDDecon.pSearchR",fitter.max_pos_search);
		Prefs.set("LSDDecon.wSearchR",fitter.max_int_search);
		double aveSL = repeatModel.REPEAT_SPACING;
		int nr = currRM.repts.length;
		if (nr > 1)
			aveSL = (currRM.repts[nr-1].x - currRM.repts[0].x)/(nr-1); // average sarcomere length
		Prefs.set("LSDDecon.aveSL", aveSL);
//		Prefs.set("LSDDecon.linescanID",currLS.linescanID);
//		Prefs.set("LSDDecon.date",currLS.date);
//		Prefs.set("LSDDecon.sample",currLS.sample);
//		Prefs.set("LSDDecon.fluor",currLS.fluor);
//		Prefs.set("LSDDecon.exposure",currLS.expTime);
		return true;
	}
*/	
/*	public SarcData[] prepSarcView() {
		double[] oy = currLS.lsInts;
//		double[] oy = this.pf;
		double[] my = currRM.calcInts;
		double[] ix = ddp.getXs(ddp.repeatResidualPixels);
		double[] xZD = ddp.getXs(currRM.repts); // X-coordinates of Z lines
		
		// Number of Sarcomeres
		int nSarc = currRM.repts.length-1;
		
		// Populate sarcomeres
		SarcData[] sd = new SarcData[nSarc];
		for (int i=0; i<nSarc; i++) {
			Point2D.Double[] szd = {currRM.repts[i],currRM.repts[i+1]};
			int iZL = (int)(float)Math.ceil(xZD[i]);
			int npix = (int)(float)Math.floor(xZD[i+1]) - iZL + 1;
			double[] xVals = new double[npix];
			double[] iObs = new double[npix];
			double[] iCalc = new double[npix];
			for (int n=0; n<npix; n++) {
				xVals[n] = ix[n+iZL]-xZD[i];
				iObs[n] = oy[n+iZL];
				iCalc[n] = my[n+iZL];
			}
			sd[i] = new SarcData(szd,xVals,iObs,iCalc);
		}
		return sd;
	}
*/	
/*	public void doExport() {
//		ddi.pStates = getParamStates(); // get the current param fix states in ddecon processor
//		ddi.lsi.linescanID = lfPFID.getText();
//		ddi.date = lfDate.getText();
//		ddi.sample = lfSample.getText();
//		ddi.fluor = lfFluor.getText();
//		try {
//			ddi.expTime = Double.parseDouble(lfExpT.getText());
//				} catch (Exception NumberFormatException) {ddi.expTime = 0;}
//		ddi.model = chModel.getSelectedItem();
//		ddi.infoDDecon = taInfo.getText();
		
		
		// Make sure results window is open and set
		if (results == null) {
			String resultsName = "Distributed Deconvolution Results";
			Frame win = WindowManager.getFrame(resultsName);
			if (win!=null){
				// set an open results window from a previous analysis.
				results = (TextWindow)win;
			}
			else {
				String defPath = "";
				resultsSavePath = Prefs.get("LSDDecon.resultsSavePath",defPath);
				// open up lastUsedResults (XXX currently set to null, call Prefs.set somewhere)
				results = openResults(resultsSavePath);
				
				// Do nothing if canceled or file does not open
				if (manager.results==null) {
		    		Dimension d = Toolkit.getDefaultToolkit().getScreenSize();
		    		int h = 200;
		    		manager.results = new TextWindow(resultsName,"",d.width,h);
		    		manager.results.setLocation(0,d.height-h-30);
		    		manager.results.addWindowListener(manager);
		    	}
			}
		}
		summarize();
		saveResults(resultsSavePath);
		
		Prefs.set("LSDDecon.gridIntvl",DDplotter.GRID_LINE_INTERVAL); // XXX not user adjustable yet
		Prefs.set("LSDDecon.autoNorm",true); // XXX Not user adjustable yet
		Prefs.set("LSDDecon.pwpp",DDplotter.PLOT_W_PERPIXEL); // XXX Not user adjustable yet
		Prefs.set("LSDDecon.minTFL",fitter.minTFL);
		Prefs.set("LSDDecon.maxTFL",fitter.maxTFL);

		Prefs.set("LSDDecon.maxOverhang",repeatModel.MAX_OVERHANG); // XXX not user adjustable yet
		Prefs.set("LSDDecon.defModIndex",currRM.cModIndex); // XXX Set to current model used
		Prefs.set("LSDDecon.pSearchR",fitter.max_pos_search);
		Prefs.set("LSDDecon.wSearchR",fitter.max_int_search);
		double aveSL = repeatModel.REPEAT_SPACING;
		int nr = currRM.repts.length;
		if (nr > 1)
			aveSL = (currRM.repts[nr-1].x - currRM.repts[0].x)/(nr-1); // average sarcomere length
		Prefs.set("LSDDecon.aveSL", aveSL);
//		Prefs.set("LSDDecon.linescanID",currLS.linescanID);
//		Prefs.set("LSDDecon.date",currLS.date);
//		Prefs.set("LSDDecon.sample",currLS.sample);
//		Prefs.set("LSDDecon.fluor",currLS.fluor);
//		Prefs.set("LSDDecon.exposure",currLS.expTime);
	}
*/	
/*	public void saveResults(String path) {
		if (path.equals("")) {
			SaveDialog sd = new SaveDialog("Save as Text", results.getTitle(), ".txt");
			String file = sd.getFileName();
			if (file == null) return;
			path = sd.getDirectory() + file;
		}
		PrintWriter pw = null;
		try {
			FileOutputStream fos = new FileOutputStream(path);
			BufferedOutputStream bos = new BufferedOutputStream(fos);
			pw = new PrintWriter(bos);
		}
		catch (IOException e) {
			//IJ.write("" + e);
			return;
		}
		results.getTextPanel().save(pw);
		pw.close();
		IJ.showStatus("");
		Prefs.set("LSDDecon.resultsSavePath",path);
	}
*/	
    /** Summarizes DDinfo for Distributed Deconvolution Results Window
     * Called by:
     * 		DDWindow.actionPerformed(ActionEvent)
     * 
     * @param appendVals: flag to output intensity values, NOT IMPLEMENTED
     */
/*    public void summarize() {
    	String ddeconHeadings = getLabelHeadingString();
    	String ddeconValues = getLabelString();
    	if (results != null) {
    		TextPanel tp = results.getTextPanel();
    		if (tp.getColumnHeadings().equals("")) {
    			tp.setColumnHeadings(" " + ddeconHeadings);;
    		}
    		int index = 1 + tp.getLineCount();
    		tp.appendLine(index + ddeconValues);
    	}
    }
*/    
    /** Returns headings for Distributed Deconvolution Results Window
     * Called by:
     * 	DDinfo.summarize(Boolean)
     * 
     * @return tab-delimited string of heading names
     */
/*    public String getLabelHeadingString() {
    	int maxZ = repeatModel.MAX_NUMBER_REPEATS;
    	String[] labels1 = {"Img FileName","Linescan ID","Date","Sample","Fluor","Exp. Time",
    			"Model","Error","GoodFit","Corr. Coeff.","Per. Area","Data Points","Z Lines"};
    	String[] Zlabels = new String [2*maxZ];
    	for (int i=0; i<maxZ; i++) {
    		Zlabels[i] = "Z" + (i+1) + " Pos\t";
    		Zlabels[i+maxZ] = "Z" + (i+1) + " Wght\t";
    	}
    	String[] labels2x;
    	if (Double.parseDouble(LSDDecon_Manager.VERSION)>=3) {
    		String[] labels2 = {"Fit Method","Data/FreeParams","Free Params","Total Params",
    			"Tolerance","Num. Iter.","Max. Iter","Restarts","Params Fixed",
    			"Int. Search Rad.","Pos. Search Rad.","Max. Overhang",
    			"LS Length","LS Width(Total)","LS angle",
    			"BG1 Width","rawLS Width","BG2 Width",
    			"BG1 Calc","rawLS Calc","BG2 Calc","LS correction",
    			"ROI X1","ROI X2","ROI Y1","ROI Y2",
    			"Img FileName","Img Directory","Img Format","Img Type",
    			"Img Frame","Img Stack Size","Img Height","Img Width",
    			"Img Pixel Depth","Img Pixel Height","Img Pixel Width",
    			"Description","SkewAngle","Fit Info","DDecon Info","LS Info","Img Info"};
    		labels2x=labels2;
    	}
    	else {
    		String[] labels2 = {"Fit Method","Data/FreeParams","Free Params","Total Params",
    			"Tolerance","Num. Iter.","Max. Iter","Restarts","Params Fixed",
    			"Int. Search Rad.","Pos. Search Rad.","Max. Overhang",
    			"LS Length","LS Width(Total)","LS angle",
    			"BG1 Width","rawLS Width","BG2 Width",
    			"BG1 Calc","rawLS Calc","BG2 Calc","LS correction",
    			"ROI X1","ROI X2","ROI Y1","ROI Y2",
    			"Img FileName","Img Directory","Img Format","Img Type",
    			"Img Frame","Img Stack Size","Img Height","Img Width",
    			"Img Pixel Depth","Img Pixel Height","Img Pixel Width",
    			"Description","Fit Info","DDecon Info","LS Info","Img Info"};
    		labels2x=labels2;
    	}
    	StringBuffer output = new StringBuffer(200);
    	output.append("\t");
    	for (int i=0; i<labels1.length; i++) {
    		output.append(labels1[i] + "\t");
    	}
    	for (int j=0; j<currRM.model.getParamNames().length; j++) {
    		output.append(currRM.model.getParamNames()[j] + "\t");
    	}
    	for (int j=0; j<2*maxZ; j++) {
    		output.append(Zlabels[j]);
    	}
    	for (int i=0; i<labels2x.length; i++) {
    		output.append(labels2x[i] + "\t");
    	}
    	return new String(output);
    }
*/    
    /** Returns values of DDinfo object for Distributed Deconvolution Results Window
     * Called by:
     * 	DDinfo.summarize(Boolean) 
     * 
     * @return tab-delimited string of DDinfo values
     */
/*    public String getLabelString() {
		int nParamsTotal = currRM.model.getParamNumber() + 2*currRM.repts.length;
		int nParamsFree = 0;
		boolean[] modelParamStates = currRM.model.getParamStates();
		boolean[][] reptParamStates = currRM.reptFixStates;
		if (modelParamStates!=null) {
			int index = 0;
			for (int i=0; i<modelParamStates.length; i++) {
				if (!modelParamStates[i]) index++;
			}
			for (int i=0; i<reptParamStates[0].length; i++) {
				if (!reptParamStates[0][i]) index++;
				if (!reptParamStates[1][i]) index++;
			}
			nParamsFree = index;
		}
//		int nParamsFree = ddi.getNumFreeParams();
		double dataParamRatio = (nParamsFree==0)?Double.NaN:((double)currLS.lsInts.length/nParamsFree);
    	int maxRepts = repeatModel.MAX_NUMBER_REPEATS;
    	String infoFit = IJ.d2s(fitter.getSumResidualsSqr(),2)+ " " + IJ.d2s(fitter.getChiSquared(),2)+ " " + fitter.info;
    	FileInfo ofi = currLS.imp.getOriginalFileInfo();
    	
    	StringBuffer sb = new StringBuffer(200);
    	sb.append("\t");
    	sb.append(ofi.fileName + "\t");
    	sb.append(ddwin.lfPFID.getText() + "\t");
    	sb.append(ddwin.lfDate.getText() + "\t");
    	sb.append(ddwin.lfSample.getText() + "\t");
    	sb.append(ddwin.lfFluor.getText() + "\t");
    	sb.append(ddwin.lfExpT.getText() + "\t");
    	sb.append(currRM.model.getTitle() + "\t");

    	sb.append(IJ.d2s(ddp.getError(),2) + "\t");
    	sb.append(IJ.d2s(fitter.getFitGoodness(),6) + "\t");
    	sb.append(IJ.d2s(ddp.getCorrel(),4) + "\t");
    	sb.append(IJ.d2s(ddp.getPerArea(),2) + "\t");
    	sb.append(currLS.lsInts.length + "\t");
    	sb.append(currRM.repts.length + "\t");

    	double[] modelParams = currRM.model.getParams();
    	for (int i=0; i<modelParams.length; i++) {
    		sb.append(IJ.d2s(modelParams[i],2) + "\t");
    	}
    	for (int i=0; i<maxRepts; i++) {
    		sb.append((i<currRM.repts.length?IJ.d2s(currRM.repts[i].x,2):"-") + "\t");
    	}
    	for (int i=0; i<maxRepts; i++) {
    		sb.append((i<currRM.repts.length?IJ.d2s(currRM.repts[i].y,2):"-") + "\t");
    	}

    	sb.append(DDCurveFitter.fitList[0] + "\t");
    	sb.append(IJ.d2s(dataParamRatio,2) + "\t");
    	sb.append(nParamsFree + "\t");
    	sb.append(nParamsTotal + "\t");
    	sb.append(fitter.maxError + "\t"); //errTol
    	sb.append(fitter.numIter + "\t");
    	sb.append(fitter.maxIter + "\t");
    	sb.append(fitter.setRestarts + "\t"); // restarts
    	sb.append(encodePStates() + "\t");
    	sb.append(fitter.max_int_search + "\t");
    	sb.append(fitter.max_pos_search + "\t");
    	sb.append(repeatModel.MAX_OVERHANG + "\t");

    	sb.append(currLS.lsLength + "\t");
    	sb.append(currLS.lsWidth + "\t");
    	sb.append(IJ.d2s(180*currLS.lsAngle/Math.PI,2) + "\t");
    	sb.append(currLS.regions[0].rect.height + "\t");
    	sb.append(currLS.regions[1].rect.height + "\t");
    	sb.append(currLS.regions[2].rect.height + "\t");
    	sb.append(currLS.regions[0].mode + "\t");
    	sb.append(currLS.regions[1].mode + "\t");
    	sb.append(currLS.regions[2].mode + "\t");
      	sb.append(LSprocessor.BG_MODES[currLS.bgCorrIndex] + "\t");

      	sb.append(IJ.d2s(currLS.line.x1d,2) + "\t"); sb.append(IJ.d2s(currLS.line.y1d,2) + "\t");
    	sb.append(IJ.d2s(currLS.line.x2d,2) + "\t"); sb.append(IJ.d2s(currLS.line.y2d,2) + "\t");

    	sb.append(ofi.fileName + "\t");
    	sb.append(ofi.directory + "\t");
        int fileFormat = ofi.fileFormat;
        String fileFormatStr = "";
    	switch (fileFormat) {
			case FileInfo.UNKNOWN: fileFormatStr = "unknown"; break;
			case FileInfo.RAW: fileFormatStr = "raw"; break;
			case FileInfo.TIFF: fileFormatStr = "tif"; break;
			case FileInfo.GIF_OR_JPG: fileFormatStr = "gif or jpg"; break;
			case FileInfo.FITS: fileFormatStr = "fits"; break;
			case FileInfo.BMP: fileFormatStr = "bmp"; break;
			case FileInfo.DICOM: fileFormatStr = "dicom"; break;
			case FileInfo.ZIP_ARCHIVE: fileFormatStr = "zip"; break;
			case FileInfo.PGM: fileFormatStr = "pgm"; break;
			default: fileFormatStr = ""; break;
    	}
    	sb.append(fileFormatStr + "\t");
    	int fileType = ofi.fileType;
    	String fileTypeStr = "";
    	switch (fileType) {
			case FileInfo.GRAY8: fileTypeStr = "byte"; break;
			case FileInfo.GRAY16_SIGNED: fileTypeStr = "short"; break;
			case FileInfo.GRAY16_UNSIGNED: fileTypeStr = "ushort"; break;
			case FileInfo.GRAY32_INT: fileTypeStr = "int"; break;
			case FileInfo.GRAY32_UNSIGNED: fileTypeStr = "uint"; break;
			case FileInfo.GRAY32_FLOAT: fileTypeStr = "float"; break;
			case FileInfo.COLOR8: fileTypeStr = "byte+lut"; break;
			case FileInfo.RGB: fileTypeStr = "RGB"; break;
			case FileInfo.RGB_PLANAR: fileTypeStr = "RGB(p)"; break;
			case FileInfo.RGB48: fileTypeStr = "RGB48"; break;
			case FileInfo.BITMAP: fileTypeStr = "bitmap"; break;
			case FileInfo.ARGB: fileTypeStr = "ARGB"; break;
			case FileInfo.BGR: fileTypeStr = "BGR"; break;
			case FileInfo.BARG: fileTypeStr = "BARG"; break;
			default: fileTypeStr = ""; break;
    	}
    	sb.append(fileTypeStr + "\t");
    	sb.append(currLS.slice + "\t");
    	sb.append(ofi.nImages + "\t");
    	sb.append(ofi.height + "\t");
    	sb.append(ofi.width + "\t");
    	sb.append(ofi.pixelDepth + "\t");
    	sb.append(ofi.pixelHeight + "\t");
    	sb.append(ofi.pixelWidth + "\t");
    	sb.append(ofi.description + "\t"); // This description contains the DDecon version
    	
    	// Append additional parameters after here
    	sb.append(IJ.d2s(currLS.skewAngle,2) + "\t");
    	
    	// Append additional parameters before here
    	sb.append(infoFit + "\t");
    	sb.append(ddwin.taInfo.getText() + "\t");
    	
//    	System.out.println("DDinfo.getLabelString: lsi.imgInfo= "+lsi.imgInfo);
//    	System.out.println("DDinfo.getLabelString: lsi.fi.info= "+lsi.fi.info);

    	int extendedHeaderIndex = ofi.info.indexOf("ExtendedHeader");
    	String shortInfo = ofi.info.substring(0,extendedHeaderIndex);
    	String imgInfo = shortInfo.replace('\n',';');
    	sb.append(lswin.taInfo.getText() + "\t");
    	sb.append(imgInfo + "\t");
//    	sb.append(lsi.infoLS + "\t"); sb.append(lsi.imp.getFileInfo().info + "\t");
//    	sb.append(lsi.infoLS + "\t"); sb.append(lsi.infoImage + "\t");
    	return new String(sb);
    }
*/    
    /** Encodes parameter states (FIXED or FREE) as a string for writing
     * Called by:
     * 	DDinfo.getLabelString() 
     * 
     * @return string of encoded parameter states
     */
/*    public String encodePStates() {
//		int nParamsTotal = currRM.model.getParamNumber() + 2*currRM.repts.length;
    	boolean[] modelParamStates = currRM.model.getParamStates();
    	int nmp = modelParamStates.length;
    	char[] f = new char [nmp];
//    	char[] f = new char [nParamsTotal];
//    	int index=0;
   		for (int j=0; j<nmp; j++) {
   			f[j] = (modelParamStates[j]) ? FIXED : FREE;
//			f[index++] = (modelParamStates[j]) ? FIXED : FREE;
   		}
    	String modelState = new String(f);
//    	String modelState = new String(f,0,nmp);
   		boolean[][] pStates = currRM.reptFixStates;
    	int nr = currRM.repts.length;
   		char[] r0 = new char[nr];
   		for (int j=0; j<nr; j++) {
   			r0[j] = (pStates[0][j]) ? FIXED : FREE;
// 			f[index++] = (pStates[0][j]) ? FIXED : FREE;
   		}
    	String posState = new String(r0);
//    	String posState = new String(f,nmp,nr);
   		char[] r1 = new char[nr];
   		for (int j=0; j<nr; j++) {
   			r1[j] = (pStates[1][j]) ? FIXED : FREE;
// 			f[index++] = (pStates[1][j]) ? FIXED : FREE;
   		}
    	String wghtState = new String(r1);
//    	String wghtState = new String(f,nmp+nr,nr);
    	
    	return modelState + "," + posState + "," + wghtState;
    }
*/    
	/** If a Distributed Deconvolution Results Window is already open, this gets it, otherwise it opens it 
	 * as a text file.
	 * Uses a routine to rebuild the heading from the first line.
	 * Called by:
	 * 		LSDDecon_Manager.actionPerformed(ActionEvent)
	 * 
	 */
/*	public TextWindow openResults(String path) {
		if (results != null) return results;
		
		Dimension d = Toolkit.getDefaultToolkit().getScreenSize();
		int h = 200;
		String name = null;
		OpenDialog od = new OpenDialog("Open Results Text File...",path);
		String directory = od.getDirectory();
		name = od.getFileName();
		if (name!=null) {
			path = directory + name;
//			YesNoCancelDialog option = new YesNoCancelDialog(this,
//					"Continue with Previous Results?","PATH="+path);
//			if (option.yesPressed()){
				results = new TextWindow(path,d.width,h);
				if (results!=null) {
					reloadResults();
		    		results.setLocation(0,d.height-h-30);
		    		results.addWindowListener(this);
					return results;
				}
//			}
		}
		String resultsName = "Distributed Deconvolution Results";
		results = new TextWindow(resultsName,"",d.width,h);
		results.getTextPanel().setColumnHeadings(" " + getLabelHeadingString());
		results.setLocation(0,d.height-h-30);
		results.addWindowListener(this);
		return results;
	}
*/	
/*	public void reloadResults() {
		TextPanel panel = results.getTextPanel();
		int lines = panel.getLineCount();
		String headings;
		String[] data = {};
		if (lines>1) {
			data = new String[lines-1];
			for (int i=0; i<lines-1; i++) {
				data[i] = panel.getLine(i+1);
			}
		}
		if (lines>0) {
			headings = panel.getLine(0);
			panel.setColumnHeadings(headings); // resets text
		}
		for (int i=0; i<data.length; i++) {
			panel.appendLine(data[i]);
		}
	}
*/	
	/** Loads a linescan from a Distributed Deconvolution Results Window.
	 * Makes a new LSinfo object from the data.
	 * Called by:
	 * 		LSDDecon_Manager.actionPerformed(ActionEvent)
	 * 
	 */
/*	void loadResultsLine() {
		// Make sure results window is open and set
		if (results == null) {
			Frame win = WindowManager.getFrame("Distributed Deconvolution Results");
			if (win!=null){
				// set an open results window from a previous analysis.
				results = (TextWindow)win;
			}
			else {
				// open up lastUsedResults (XXX currently set to null, call Prefs.set somewhere)
				results = openResults(resultsSavePath);
				
				// Do nothing if canceled or file does not open
				if (results==null) return;
			}
		}

		// Load line dialog
		GenericDialog gd = new GenericDialog("Load Results");
		gd.addNumericField("Record Number",0,0);
		gd.showDialog();
		int index = (int)gd.getNextNumber()-1;
		
		// Construct a new LSinfo object from the results and add it to the Line List.
	    	String text = results.getTextPanel().getLine(index);
	    	FileInfo ofi = new FileInfo();
	    	String[] fields = Tools.split(text,"\t");
	    	int nFields = fields.length;
	    	
	    	for (int i=0; i<nFields; i++) {
//	    		System.out.println("LSinfo(String).<init>: field["+i+"]= "+fields[i]);
	    		IJ.write("LSinfo(String).<init>: field["+i+"]= "+fields[i]);
	    	}

	    	String linescanID = "";
	    	Line line = null;
	    	if (nFields>=83) {
	    		ofi.fileName = fields[1];
	    		linescanID = fields[2];
	        	int index2 = 69;//70;
	    		int lsLength = Integer.parseInt(fields[index2++]);
	    		int lsWidth = Integer.parseInt(fields[index2++]);
	    		double lsAngle = Double.parseDouble(fields[index2++])*Math.PI/180;
	    		int bg1w = Integer.parseInt(fields[index2++]);
	    		int rawLSw = Integer.parseInt(fields[index2++]);
	    		int bg2w = Integer.parseInt(fields[index2++]);
	    		int[] regionCalcs = new int [3];
	    		for (int i=0; i<3; i++) {
	    			String calcMode = fields[index2++];
	    			if (calcMode.equalsIgnoreCase("Minimum")) regionCalcs[i] = 1;
	    			if (calcMode.equalsIgnoreCase("Average")) regionCalcs[i] = 2;
	    			if (calcMode.equalsIgnoreCase("Median")) regionCalcs[i] = 4;
	    			if (calcMode.equalsIgnoreCase("Maximum")) regionCalcs[i] = 8;
	    		}
	    		int bg1calc = regionCalcs[0];
	    		int rawLScalc = regionCalcs[1];
	    		int bg2calc = regionCalcs[2];
	    		String corrcalcStr = fields[index2++];
	    		int corrcalc = 2;
//	    		corrcalc = 2; //XXX Prefs.set(corrcalc)?
	    		if (corrcalcStr!=null) {
		    		for (int i=0; i<LSprocessor.BG_MODES.length; i++) {
		    			if (corrcalcStr.equals(LSprocessor.BG_MODES[i])) corrcalc = i;
		    		}
	    		}
	    		else {
	    			corrcalcStr.equals(LSprocessor.BG_MODES[corrcalc]);
	    		}
	    		double x1d = Double.parseDouble(fields[index2++]);
	    		double y1d = Double.parseDouble(fields[index2++]);
	    		double x2d = Double.parseDouble(fields[index2++]);
	    		double y2d = Double.parseDouble(fields[index2++]);
	        	line = new Line((int)x1d,(int)y1d,(int)x2d,(int)y2d);
	    		line.x1d = x1d;
	    		line.y1d = y1d;
	    		line.x2d = x2d;
	    		line.y2d = y2d;
	    		ofi.fileName = fields[84];
	    		ofi.directory = fields[85];
	    		ofi.fileFormat = FileInfo.GRAY16_SIGNED; //fields[86]; XXX Update
	    		ofi.fileType = FileInfo.RAW; //fields[87]; XXX Update
	    		
	    		index2 = 87;//88;
	    		int slice = Integer.parseInt(fields[index2++]);
	    		ofi.nImages = Integer.parseInt(fields[index2++]);
	    		ofi.height = Integer.parseInt(fields[index2++]);
	    		ofi.width = Integer.parseInt(fields[index2++]);
	    		ofi.pixelDepth = Double.parseDouble(fields[index2++]);
	    		ofi.pixelHeight = Double.parseDouble(fields[index2++]);
	    		ofi.pixelWidth = Double.parseDouble(fields[index2++]);
	    		ofi.description = fields[index2++];
	    		int n = ofi.description.indexOf("LSDDecon v");
	    		String s = "";
				if (n>0) s = ofi.description.substring(n+10,n+14);
				System.out.println("LSinfo.<init>: s= "+s);
	    		double version;
	    			try {
	    				version = Double.parseDouble(s);
	    			}
	    			catch (NumberFormatException e) {
	    				version = 0;
	    			}
	    			
	    		// Additional fields to add here
	    		double skewAngle = 0;
	    		if (version>=3) {
	    			skewAngle = Double.parseDouble(fields[index2++]);
	    		}
	    		
	    		String info = "";
	    		for (int i=index2; i<nFields; i++) {
	        		info = info + "; " + fields[index2++];
	    		}
	    		
	    		String fitInfo = fields[index++];
	    		String ddeconInfo = fields[index++];
	    		this.infoLS = fields[index++];
	    		fi.info = fields[index];
	    		
	    		
	    	}
//		LSinfo lsi = new LSinfo(results.getTextPanel().getLine(index));
//		IJ.write("Manager.loadResultsLine: lsi="+lsi.toString());
//		lsi.manager = this;
		String label = this.getUniqueLSName(linescanID);
//		String label = this.getUniqueLSName(lsi.linescanID);
		list.add(label);
		rois.put(label,line);
		info.setText(getInfoText(label));
		list.select(list.getItemCount());
	}
*/	
/*    public void loadSubImage(String name) {
    	String lsIDprefix = this.prefixField.getText();
		String linescanID = getUniqueLSName(lsIDprefix+"-1"); // Assign new linescanID
    	String basename = currLS.imp.getShortTitle()+"_"+currLS.slice+"_"+linescanID;
    	String defName = currLS.imp.getOriginalFileInfo().fileName + "_" + currLS.slice + "_" + linescanID + ".tif";
    	String dir = currLS.imp.getOriginalFileInfo().directory;
    	if (name.equals("")) {
    		OpenDialog od = new OpenDialog("Open TIFF subimage",dir,defName);
    		dir = od.getDirectory();
    		name = od.getFileName();
    	}
    	Opener o = new Opener();
    	currLS.subImage = o.openTiff(dir,name);
    }
*/    
	/** Adds active line from image to list with options.
	 * Called by:
	 * 		LSDDecon_Manager.actionPerformed(ActionEvent)
	 * 		LSDDecon_Manager.runCommand(String)
	 * 
	 * @param shiftKeyDown: calls addAndDraw(boolean).
	 * @param altKeyDown: flag for calling add(boolean).
	 * @see add(boolean)
	 * @see addAndDraw(boolean)
	 */
	// NOT CURRENTLY USED
	void add(boolean shiftKeyDown, boolean altKeyDown) {
		if (altKeyDown)	add(true); // Prompts for name
		else add(false); // Automatically assigns name
	}

	/** Adds active line from image to list.
	 * Called by:
	 * 		LSDDecon_Manager.add(boolean,boolean)
	 * 		LSDDecon_Manager.addAndDraw(boolean)
	 * 
	 * @param promptForName
	 * @return flag for whether image has a line selection
	 */
	boolean add(boolean promptForName) {
		ImagePlus imp = getImage();
		if (imp==null)
			return false;
		Roi roi = imp.getRoi();
		if ((roi==null) || !(roi instanceof Line)) {
			error("The active image does not have a line selection.");
			return false;
		}
		Line line = (Line) roi;
		String name = line.getName();
//		int slice1 = imp.getCurrentSlice();
		String label = name!=null?name:getLSName(imp);
//		String label = name!=null?name:(prefixField.getText()+"-1");
//		String label = name!=null?name:getLabel(imp, line);
		if (promptForName)
			label = getName(label);
		else
			label = getUniqueName(label);
		if (label==null) return false;
		int index = list.getItemCount();
		list.add(label);
		line.setName(label);
		lineCopy = (Line)line.clone();
//		slice2 = slice1;
		LineRegion lr = new LineRegion(line);
//		LineRegion lr = new LineRegion(lineCopy);
		rois.put(label, lr);
//		rois.put(label, lineCopy);
		String text = getInfoText(label);
		info.setText(text);
		list.select(index);
		if (Recorder.record) Recorder.record("roiManager", "Add");
		return true;
	}
	
	public String getLSName (ImagePlus imp) {
		String name = prefixField.getText()+"-1";
//		String name = imp.getShortTitle() + nameDelimiter + prefixField.getText()+"-1";
//		int slice = imp.getCurrentSlice();
//		String name = imp.getShortTitle() + nameDelimiter + slice + nameDelimiter + prefixField.getText()+"-1";
		return name;
	}
	
	/** Gets a label for the image Roi based on Roi bounds and image slice
	 * Called by:
	 * 		LSDDecon_Manager.add(boolean)
	 * 
	 * @param imp: image which active line is on
	 * @param roi: active line
	 * @return
	 */
	String getLabel(ImagePlus imp, Roi roi) {
		Rectangle r = roi.getBounds();
		int xc = r.x + r.width/2;
		int yc = r.y + r.height/2;
		if (xc<0) xc = 0;
		if (yc<0) yc = 0;
		int digits = 4;
		String xs = "" + xc;
		if (xs.length()>digits) digits = xs.length();
		String ys = "" + yc;
		if (ys.length()>digits) digits = ys.length();
		xs = "000" + xc;
		ys = "000" + yc;
		String label = ys.substring(ys.length()-digits) + "-" + xs.substring(xs.length()-digits);
		if (imp.getStackSize()>1) {
			String zs = "000" + imp.getCurrentSlice();
			label = zs.substring(zs.length()-digits) + "-" + label;
		}
		return label;
	}

	/** Replaces a selected line entry with the line on the selected or active image.  This
	 * is normally automatic in the RoiManager class, but I preferred to have to push this button.
	 * Called by:
	 * 		LSDDecon_Manager.actionPerformed(ActionEvent)
	 * 
	 */
	boolean applyToImage() {
		int index = list.getSelectedIndex();
		if (index<0)
			return error("Exactly one item in the list must be selected.");
		boolean setSlice = false;
		String label = list.getItem(index);
		Line line;
		line = getLineFromHashTable(label);
		ImagePlus imp = getImage();
		if (imp==null || line==null) {
			return false;
		}
        if (setSlice) {
            int slice = getSliceNumber(label);
            if (slice>=1 && slice<=imp.getStackSize())
                imp.setSlice(slice);
        }
		imp.setRoi(line);
//		imp.setRoi((Line)line.clone());
		return true;
	}
	
	/** Renames the current line or LSinfo object in the list.
	 * Called by:
	 * 		LSDDecon_Manager.actionPerformed(ActionEvent)
	 * 
	 */
	boolean rename() {
		int index = list.getSelectedIndex();
		if (index<0)
			return error("Exactly one item in the list must be selected.");
		String name = list.getItem(index);
		String name2 = getName(name);
		if (name2==null) return false;
		LineRegion lr = (LineRegion)rois.get(name);
//		Object o = rois.get(name);
//		if (o instanceof Line) {
//		Line line = (Line)o;
		rois.remove(name);
		lr.line.setName(name2);
		rois.put(name2,lr);
//		line.setName(name2);
//		rois.put(name2, line);
//		}
//		else {
//			LSinfo lsi = (LSinfo)o;
//			rois.remove(name);
//			lsi.linescanID = name2;
//			rois.put(name2, lsi);
//		}
		list.replaceItem(name2, index);
		list.select(index);
		return true;
	}
	
	boolean deselect() {
		int count = list.getItemCount();
		for (int i=0; i<count; i++) {
			list.deselect(i);
		}
		return true;
	}
	
	/** Deletes selected or all items.
	 * Called by:
	 * 		LSDDecon_Manager.actionPerformed(ActionEvent)
	 * 		LSDDecon_Manager.runCommand(String)
	 * 
	 * @param replacing
	 */
	boolean delete(boolean replacing) {
		int count = list.getItemCount();
		if (count==0)
			return error("The list is empty.");
		int index[] = list.getSelectedIndexes();
		if (index.length==0 || (replacing&&count>1)) {
			String msg = "Delete all items on the list?";
			if (replacing)
				msg = "Replace items on the list?";
			canceled = false;
			if (!IJ.macroRunning() && !macro) {
				YesNoCancelDialog d = new YesNoCancelDialog(this, "ROI Manager", msg);
				if (d.cancelPressed())
					{canceled = true; return false;}
				if (!d.yesPressed()) return false;
			}
			index = getAllIndexes();
		}
		for (int i=count-1; i>=0; i--) {
			boolean delete = false;
			for (int j=0; j<index.length; j++) {
				if (index[j]==i)
					delete = true;
			}
			if (delete) {
				rois.remove(list.getItem(i));
				list.remove(i);
			}
		}
		return true;
	}
	
	/** Displays a dialog to enter a name.  Returns a unique name.
	 * Called by:
	 * 		LSDDecon_Manager.add(boolean)
	 * 		LSDDecon_Manager.rename()
	 * 
	 * @param name: supplied name.
	 * @return unique version of the name.
	 */
	String getName(String name) {
		GenericDialog gd = new GenericDialog("LSDDecon Manager");
		gd.addStringField("Rename As:", name, 20);
		gd.showDialog();
		if (gd.wasCanceled())
			return null;
		String name2 = gd.getNextString();
		name2 = getUniqueName(name2);
		return name2;
	}

	/** Restores a line or Lsinfo line to an image.  LSinfo line is made and subpixel coordinates are set
	 *  from the LSinfo data.
	 *  
	 *  Called by:
	 *  	LSDDecon_Manager.draw()
	 *  	LSDDecon_Manager.measure()
	 *  	LSDDecon_Manager.save()
	 *  	LSDDecon_Manager.select(int)
	 *  */
	boolean restore(int index, boolean setSlice) {
		String label = list.getItem(index);
		Line line;
		line = getLineFromHashTable(label);
		ImagePlus imp = getImage();
		if (imp==null || line==null)
			return false;
        if (setSlice) {
            int slice = getSliceNumber(label);
            if (slice>=1 && slice<=imp.getStackSize())
                imp.setSlice(slice);
        }
		imp.setRoi((Roi)line.clone());
		return true;
	}
	
	/**
	 * Called by:
	 * 		LSDDecon_Manager.applyToImage()
	 * 		LSDDecon_Manager.measure()
	 * 		LSDDecon_Manager.restore(int,boolean)
	 * @param label
	 * @return
	 */
	int getSliceNumber(String label) {
		int slice = -1;
		if (label.length()>4 && label.charAt(4)=='-' && label.length()>=14)
			slice = (int)Tools.parseDouble(label.substring(0,4),-1);
		return slice;
	}
	
	/** Only ROIs can be opened, not LSinfo objects
	 * Called by:
	 * 		LSDDecon_Manager.actionPerformed(ActionEvent)
	 * 		LSDDecon_Manager.runCommand(String)
	 * 
	 * @param path
	 */
	void open(String path) {
		Macro.setOptions(null);
		String name = null;
		if (path==null) {
			OpenDialog od = new OpenDialog("Open Selection(s)...", "");
			String directory = od.getDirectory();
			name = od.getFileName();
			if (name==null)
				return;
			path = directory + name;
		}
		if (Recorder.record) Recorder.record("roiManager", "Open", path);
		if (path.endsWith(".zip")) {
			openZip(path);
			return;
		}
		Opener o = new Opener();
		if (name==null) name = o.getName(path);
		Roi roi = o.openRoi(path);
		if (roi!=null) {
			if (name.endsWith(".roi"))
				name = name.substring(0, name.length()-4);
			name = getUniqueName(name);
			list.add(name);
			rois.put(name, roi);
		}		
	}
	
	// Modified on 2005/11/15 by Ulrik Stervbo to only read .roi files and to not empty the current list
	/** 
	 * Called by:
	 * 		LSDDecon_Manager.open(String)
	 * 
	 * @param path
	 */
	void openZip(String path) { 
		ZipInputStream in = null; 
		ByteArrayOutputStream out; 
		int nRois = 0; 
		try { 
			in = new ZipInputStream(new FileInputStream(path)); 
			byte[] buf = new byte[1024]; 
			int len; 
			ZipEntry entry = in.getNextEntry(); 
			while (entry!=null) { 
				String name = entry.getName(); 
				if (name.endsWith(".roi")) { 
					out = new ByteArrayOutputStream(); 
					while ((len = in.read(buf)) > 0) 
						out.write(buf, 0, len); 
					out.close(); 
					byte[] bytes = out.toByteArray(); 
					RoiDecoder rd = new RoiDecoder(bytes, name); 
					Roi roi = rd.getRoi(); 
					if (roi!=null) { 
						name = name.substring(0, name.length()-4); 
						name = getUniqueName(name); 
						list.add(name); 
						rois.put(name, roi); 
						nRois++;
					} 
				} 
				entry = in.getNextEntry(); 
			} 
			in.close(); 
		} catch (IOException e) {error(e.toString());} 
		if(nRois==0)
				error("This ZIP archive does not appear to contain \".roi\" files");
	} 

	/** 
	 * Called by:
	 * 		LSDDecon_Manager.applyToImage()
	 * 		LSDDecon_Manager.restore(int,boolean)
	 * 		LSDDecon_Manager.saveMultiple(int[],String)
	 * 
	 * @param label
	 */
	Line getLineFromHashTable(String label) {
		Object o = rois.get(label);
		LineRegion lr = (LineRegion)o;
		Line line = lr.line;
//		Line line;
//		if (o instanceof Line) {
//			line = (Line)o;
//		}
//		else if (o instanceof LSinfo){
//			LSinfo lsi = (LSinfo)o;
//			line = lsi.line;
//		}
//		else {
//			line = null;
//			IJ.error("LSDDecon Manager","A line cannot be derived from this selection!");
//		}
		return line;
	}
	
	/** 
	 * Called by:
	 * 		LSDDecon_Manager.add(boolean)
	 * 		LSDDecon_Manager.getName(String)
	 * 		LSDDecon_Manager.open(String)
	 * 		LSDDecon_Manager.openZip(String)
	 * 
	 * @param name
	 */
	String getUniqueName(String name) {
		String name2 = name;
		int n = 1;
		Object o = rois.get(name2);
		while (o!=null) {
			o = rois.get(name2);
			if (o!=null) {
				int lastDash = name2.lastIndexOf("-");
				if (lastDash!=-1 && name2.length()-lastDash<5)
					name2 = name2.substring(0, lastDash);
				name2 = name2+"-"+n;
				n++;
			}
			o = rois.get(name2);
		}
		return name2;
	}

	/** 
	 * Called by:
	 * 		LSDDecon_Manager.getInfo(String)
	 * 		LSDDecon_Manager.loadResultsLine()
	 * 
	 * @param name
	 */
/*	String getUniqueLSName(String name) {
			String name2 = name;
			int n = 1;
			Line roi2 = (Line)rois.get(name2);
//			LSinfo roi2 = (LSinfo)rois.get(name2);
			while (roi2!=null) {
					int lastDash = name2.lastIndexOf("-");
					if (lastDash!=-1 && name2.length()-lastDash<5)
						name2 = name2.substring(0, lastDash);
					name2 = name2+"-"+n;
					n++;
					roi2 = (Line)rois.get(name2);
//				roi2 = (LSinfo)rois.get(name2);
			}
			return name2;
	}
*/	
	/** Only works with Lines, not LSinfo objects.
	 * Called by:
	 * 		LSDDecon_Manager.actionPerformed(ActionEvent)
	 * 
	 */
	boolean save() {
		if (list.getItemCount()==0)
			return error("The selection list is empty.");
		int[] indexes = list.getSelectedIndexes();
		if (indexes.length==0)
			indexes = getAllIndexes();
		if (indexes.length>1)
			return saveMultiple(indexes, null);
		String name = list.getItem(indexes[0]);
		Macro.setOptions(null);
		SaveDialog sd = new SaveDialog("Save Selection...", name, ".roi");
		String name2 = sd.getFileName();
		if (name2 == null)
			return false;
		String dir = sd.getDirectory();
		Object o = rois.get(name);
		Line line;
//		if (o instanceof Line) {
			line = (Line)o;
//		}
//		else if (o instanceof LSinfo) {
//			LSinfo lsi = (LSinfo)o;
//			line = lsi.line;
//		}
//		else {
//			IJ.error("LSDDecon Manager","Item not saved, line infomation not available");
//			return false;
//		}
		rois.remove(name);
		if (!name2.endsWith(".roi")) name2 = name2+".roi";
		String newName = name2.substring(0, name2.length()-4);
		rois.put(newName, line);
		line.setName(newName);
		list.replaceItem(newName, indexes[0]);
		if (restore(indexes[0], true))
			IJ.run("Selection...", "path='"+dir+name2+"'");
		return true;
	}

	/** 
	 * Called by:
	 * 		LSDDecon_Manager.runCommand(String,String)
	 * 		LSDDecon_Manager.save()
	 * 
	 */
	boolean saveMultiple(int[] indexes, String path) {
		Macro.setOptions(null);
		if (path==null) {
			SaveDialog sd = new SaveDialog("Save ROIs...", "RoiSet", ".zip");
			String name = sd.getFileName();
			if (name == null)
				return false;
			String dir = sd.getDirectory();
			path = dir+name;
		}
		try {
			ZipOutputStream zos = new ZipOutputStream(new FileOutputStream(path));
			DataOutputStream out = new DataOutputStream(new BufferedOutputStream(zos));
			RoiEncoder re = new RoiEncoder(out);
			for (int i=0; i<indexes.length; i++) {
				String label = list.getItem(indexes[i]);
				Line line;
				line = this.getLineFromHashTable(label);
				if (!label.endsWith(".line")) label += ".line";
        		zos.putNextEntry(new ZipEntry(label));
				re.write(line);
				out.flush();
			}
			out.close();
		}
		catch (IOException e) {
			error(""+e);
			return false;
		}
		if (Recorder.record) Recorder.record("roiManager", "Save", path);
		return true;
	}
	
	/** 
	 * Called by:
	 * 		LSDDecon_Manager.actionPerformed(ActionEvent)
	 * 		LSDDecon_Manager.runCommand(String,String)
	 * 
	 */
/*	boolean measure() {
		ImagePlus imp = getImage();
		if (imp==null)
			return false;
		int[] indexes = list.getSelectedIndexes();
		if (indexes.length==0)
			indexes = getAllIndexes();
        if (indexes.length==0) return false;

		int nLines = 0;
		for (int i=0; i<indexes.length; i++) {
			String label = list.getItem(indexes[i]);
			Roi roi = (Roi)rois.get(label);
			if (roi.isLine()) nLines++;
		}
		if (nLines>0 && nLines!=indexes.length) {
			error("All items must be areas or all must be lines.");
			return false;
		}
						
		int nSlices = 1;
		String label = list.getItem(indexes[0]);
		if (getSliceNumber(label)==-1 || indexes.length==1) {
			int setup = IJ.setupDialog(imp, 0);
			if (setup==PlugInFilter.DONE)
				return false;
			nSlices = setup==PlugInFilter.DOES_STACKS?imp.getStackSize():1;
		}
		int currentSlice = imp.getCurrentSlice();
		for (int slice=1; slice<=nSlices; slice++) {
			if (nSlices>1) imp.setSlice(slice);
			for (int i=0; i<indexes.length; i++) {
				if (restore(indexes[i], nSlices==1))
					IJ.run("Measure");
				else
					break;
			}
		}
		imp.setSlice(currentSlice);
		if (indexes.length>1)
			IJ.run("Select None");
		if (Recorder.record) Recorder.record("roiManager", "Measure");
		return true;
	}	
*/
	/** 
	 * Called by:
	 * 		LSDDecon_Manager.actionPerformed(ActionEvent)
	 * 		LSDDecon_Manager.runCommand(String,String)
	 * 
	 */
/*	boolean draw() {
		int[] indexes = list.getSelectedIndexes();
		if (indexes.length==0)
			indexes = getAllIndexes();
		ImagePlus imp = WindowManager.getCurrentImage();
		Undo.setup(Undo.COMPOUND_FILTER, imp);
		for (int i=0; i<indexes.length; i++) {
			if (restore(indexes[i], true)) {
				IJ.run("Draw");
				IJ.run("Select None");
			} else
				break;
		}
		Undo.setup(Undo.COMPOUND_FILTER_DONE, imp);
		if (Recorder.record) Recorder.record("roiManager", "Draw");
		return true;
	}
*/
	/** 
	 * Called by:
	 * 		LSDDecon_Manager.actionPerformed(ActionEvent)
	 * 		LSDDecon_Manager.runCommand(String,String)
	 * 
	 */
/*	void combine() {
		ImagePlus imp = getImage();
		if (imp==null) return;
		int[] indexes = list.getSelectedIndexes();
		if (indexes.length<=1)
			indexes = getAllIndexes();
		ShapeRoi s1=null, s2=null;
		for (int i=0; i<indexes.length; i++) {
			Roi roi = (Roi)rois.get(list.getItem(indexes[i]));
			if (roi.isLine() || roi.getType()==Roi.POINT)
				continue;
			if (s1==null) {
				if (roi instanceof ShapeRoi)
					s1 = (ShapeRoi)roi;
				else
					s1 = new ShapeRoi(roi);
				if (s1==null) return;
			} else {
				if (roi instanceof ShapeRoi)
					s2 = (ShapeRoi)roi;
				else
					s2 = new ShapeRoi(roi);
				if (s2==null) continue;
				if (roi.isArea())
					s1.or(s2);
			}
		}
		if (s1!=null)
			imp.setRoi(s1);
		if (Recorder.record) Recorder.record("roiManager", "Combine");
	}
*/
	int[] getAllIndexes() {
		int count = list.getItemCount();
		int[] indexes = new int[count];
		for (int i=0; i<count; i++)
			indexes[i] = i;
		return indexes;
	}

	/** Returns the selected image, the current image if none is selected, or null if there are no images.
	 * Called by:
	 * 		LSDDecon_Manager.add(boolean)
	 * 		LSDDecon_Manager.applyToImage()
	 * 		LSDDecon_Manager.combine()
	 * 		LSDDecon_Manager.doLineScan()
	 * 		LSDDecon_Manager.getInfo(String)
	 * 		LSDDecon_Manager.measure()
	 * 		LSDDecon_Manager.restore(int,boolean)
	 * 		LSDDecon_Manager.update()
	 * 
	 */
	ImagePlus getImage() {		
		ImagePlus imp;
		int index = windowChoice.getSelectedIndex();
		if (index>=0) {
			int winID = this.imgID[index];
			imp = WindowManager.getImage(winID);
		}
		else 
			imp = WindowManager.getCurrentImage();
		
		if (imp==null) {
			return null;
		} else
			return imp;
	}

	boolean error(String msg) {
		new MessageDialog(this, "ROI Manager", msg);
		Macro.abort();
		return false;
	}
	
/*	public void processWindowEvent(WindowEvent e) {
		super.processWindowEvent(e);
		if (e.getID()==WindowEvent.WINDOW_CLOSING) {
			instance = null;	
		}
		ignoreInterrupts = false;
	}
*/	
	/** Returns a reference to the ROI Manager
		or null if it is not open. NOT CALLED*/
/*	public static LSDDecon_Manager getInstance() {
		return (LSDDecon_Manager)instance;
	}
*/
	/** Returns the ROI Hashtable. NOT CALLED*/
/*	public Hashtable getROIs() {
		return rois;
	}
*/
	/** Returns the selection list. NOT CALLED*/
/*	public List getList() {
		return list;
	}
*/		
	/** 
	 * Called by:
	 * 		LSDDecon_Manager.actionPerformed(ActionEvent)
	 * 		LSDDecon_Manager.runCommand(String,String)
	 * 
	 */
/*	public void select(int index) {
		int n = list.getItemCount();
		if (index<0) {
			for (int i=0; i<n; i++)
				if (list.isSelected(i)) list.deselect(i);
			return;
		}
		boolean mm = list.isMultipleMode();
		if (mm) list.setMultipleMode(false);
		if (index<n) {
			list.select(index);
			restore(index, true);	
			if (!Interpreter.isBatchMode())
				IJ.wait(10);
		}
		if (mm) list.setMultipleMode(true);
	}
*/	
/*	public static String getVersion() {
		return VERSION;
	}
*/
    /** Overrides PlugInFrame.close(). */
    public void close() {
    	Prefs.set("LSDDecon.lsprefix",prefixField.getText());
    	super.close();
    	instance = null;
    }
}

