/*  DDecon - Distributed Deconvolution Plugin for ImageJ
    Copyright (C) 2007  Ryan Littlefield

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

import java.awt.Checkbox;
import java.awt.Label;
import java.awt.Panel;
import java.awt.Rectangle;
import java.awt.event.ItemListener;

import ij.IJ;
import ij.ImagePlus;
import ij.process.ImageProcessor;

public class LSRegion {
	public static final int MIN = -1;
	public static final int AVE = -2;
	public static final int MED = -4;
	public static final int MAX = -8;
	public static final int[] MODES = {MAX,MED,AVE,MIN};
	
	public int mode; // Negative values get MIN, AVE, MED, or MAX; positive values give a constant
	public double[] ints;
	public Rectangle rect;
	public Panel p;
	public Label label;
	public Checkbox [] views;
	public String name;
	
	public LSRegion(Rectangle rect, int mode) {
		this.rect = rect;
		ints = new double[rect.width];
		this.mode = mode;
	}
	
	public int getViewMode() {
		int mode = 0;
    	for (int j=0; j<LSRegion.MODES.length; j++) {
			mode += (views[j].getState()?LSRegion.MODES[j]:0);
		}
    	return mode;
	}
	
	public double[] calculateIntensities(ImagePlus imp, int mode) {
		ImageProcessor ip = imp.getProcessor();
		int len = ip.getWidth();
		double[] ints = new double[len];
		double[] column = new double[rect.height];
		for (int i=0; i<rect.width; i++) {
			for (int j=0; j<rect.height; j++) {
				column[j] = ip.getPixelValue(i+rect.x,j+rect.y);
			}
			if (mode<0) {
				switch (mode) {
				case MIN: 
					ints[i] = getMinimum(column);
					break;
				case AVE: 
					ints[i] = getAverage(column);
					break;
				case MED: 
					ints[i] = getMedian(column);
					break;
				case MAX: 
					ints[i] = getMaximum(column);
					break;
				}
			}
			else ints[i] = mode;
		}
		return ints;
	}
	
/*	public double[] calculateIntensities(ImageProcessor ip, int mode) {
//		ImageProcessor ip = imp.getProcessor();
		int len = ip.getWidth();
		double[] ints = new double[len];
		double[] column = new double[rect.height];
		for (int i=0; i<rect.width; i++) {
			for (int j=0; j<rect.height; j++) {
				column[j] = ip.getPixelValue(i+rect.x,j+rect.y);
			}
			if (mode<0) {
				switch (mode) {
				case MIN: 
					ints[i] = getMinimum(column);
					break;
				case AVE: 
					ints[i] = getAverage(column);
					break;
				case MED: 
					ints[i] = getMedian(column);
					break;
				case MAX: 
					ints[i] = getMaximum(column);
					break;
				}
			}
			else ints[i] = mode;
		}
		return ints;
	}
*/	
	public void makePanel(int w, int h, int n) {
		p = new Panel();
		p.setLayout(null);
		p.setSize(w,h);
		int hc = (h/(n+1));
		label = new Label(name);
		label.setSize(w,(int)hc);
		label.setLocation(0,0);
		p.add(label);
		views = new Checkbox [n];
		for (int i=0;i<views.length;i++) {
			views[i] = new Checkbox("",null,false);
			views[i].setSize(w,(int)hc);
			views[i].setLocation(0,(int)((i+1)*hc));
			p.add(views[i]);
		}
	}
	
	public Panel getPanel() {
		return p;
	}
	
	public void addItemListener(ItemListener arg0) {
		for (int i=0;i<views.length;i++) views[i].addItemListener(arg0);
	}
	
	public static String getModeString(int mode) {
		String s = "("+IJ.d2s(mode,0)+")";
		switch (mode) {
		case MIN: s = "Minimum"; break;
		case AVE: s = "Average"; break;
		case MED: s = "Median"; break;
		case MAX: s = "Maximum"; break;
		}
		return s;
	}
	
	private static double getMinimum(double[] vals) {
		double min = vals[0];
		for (int i=1; i<vals.length; i++) min = Math.min(min,vals[i]);
		return min;
	}
	
	private static double getAverage(double[] vals) {
		double ave = 0;
		double sum = 0;
		for (int i=0; i<vals.length; i++) sum += vals[i];
		ave = sum/vals.length;
		return ave;
	}

	private static double getMedian(double[] vals) {
		// Assuming list is small, the first half is sorted
		int len = vals.length;
		int medInd = (len % 2!=0) ? (len-1)/2 : len/2; //medInd = index of median in the sorted list (or second of 2)
		for (int i=0; i<=medInd;i++) {	// index i is the position looking to be filled in the sorted list
			double min = vals[i];
			int minInd = i;
			for (int j=i+1; j<len; j++) {	// index j is the position being searched to fill the looked-for position
				if (vals[j]<min) {
					minInd = j;	// by the end of the search minInd is the index of the minimum value
					min = vals[j];	// and min, is that value
				}
			}
			// swap vals[minInd] with vals[i]
			double temp = vals[i];
			vals[i] = vals[minInd];
			vals[minInd] = temp;
		}
		double med = (len % 2!=0) ? vals[medInd] : (vals[medInd-1]+vals[medInd])/2;	// take middle or ave middle 2
		return med;
	}
	
	private static double getMaximum(double[] vals) {
		double max = 0;
		for (int i=1; i<vals.length; i++) max = Math.max(max,vals[i]);
		return max;
	}
}
