/*  DDecon - Distributed Deconvolution Plugin for ImageJ
    Copyright (C) 2007  Ryan Littlefield

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

import java.awt.geom.*;

abstract class SModel {
	static final int MIN_LENGTH = 31;
	private String modelTitle;
	private String modelDesc;
	private String[] paramNames;
	private double[] params;
	private boolean[] isFixed;
	private boolean[] isBoundedLo;
	private boolean[] isBoundedHi;
	private double[] boundsLo;
	private double[] boundsHi;
	private double[] simplexSteps;
	PSFModel psf = null;
//	public double psfw = 2;
//	public double tflen = 8;
//	public double tfuni = 0;
//	public double tfpend = 0;
//	public double tfbend = 1;
//	public boolean [] fParams= {false, false, false, false, true};
	
	SModel(String modelTitle) {this.modelTitle = modelTitle;}

	SModel(String title, String desc) {
		this.modelTitle = title;
		this.modelDesc = desc;
	}
	
	SModel(String title, String desc, double[] params, boolean[] states) {
		this.modelTitle = title;
		this.modelDesc = desc;
		this.params = params;
		this.isFixed = states;
	}
	
//	SModel(String modelTitle, double [] dParams, boolean[] fParams) {
//		this.modelTitle = modelTitle;
//		psfw = dParams[0]; tflen = dParams[1]; tfuni = dParams[2]; tfpend = dParams[3];
//		if (dParams.length>4) tfbend = dParams[4];
//		this.fParams = fParams;
//	}
	
	public String getTitle() {return modelTitle;}
	public String getDescript() {return modelDesc;}
	public int getParamNumber() {return params.length;}

	public String[] getParamNames() {return paramNames;}
	public double[] getSimplexSteps() {return simplexSteps;}
	public double[] getParams() {
		double[] p = params;
		if (psf!=null) p[0] = psf.focusParam;
		return p;
	}
	public boolean[] getParamStates() {
		boolean[] s = isFixed;
		if (psf!=null) s[0] = psf.pState;
		return s;
	}
	public boolean[] getIsBoundLo() {return isBoundedLo;}
	public boolean[] getIsBoundHi() {return isBoundedHi;}
	public double[] getBoundsLo() {return boundsLo;}
	public double[] getBoundsHi() {return boundsHi;}
	
	public void setParamNames(String[] names) {paramNames = names;}
	public void setSimplexSteps(double[] steps) {simplexSteps = steps;}
	public void setParams(double[] p) {
		params = p;
		if (psf!=null) psf.focusParam = p[0];
	}
	public void setParamStates(boolean[] s) {
		isFixed = s;
		if (psf!=null) psf.pState = s[0];
	}
	public void setIsBoundLo(boolean[] b) {isBoundedLo = b;}
	public void setIsBoundHi(boolean[] b) {isBoundedHi = b;}
	public void setBoundsLo(double[] d) {boundsLo = d;}
	public void setBoundsHi(double[] d) {boundsHi = d;}
	
	public double[] encodeParamValues(double[] params, boolean[] isFixed, double[] boundsLo, double[] boundsHi) {
//		for (int i=0; i<isFixed.length; i++) System.out.println("SModel.encodeParamValues: isFixed["+i+"]="+isFixed[i]);
		int free = 0;
//		System.out.println("SModel.encodeParamValues: params.length="+params.length);
		for (int i=0; i<params.length; i++) if (!isFixed[i]) free++;
		double [] eParams = new double [free];
		int index = 0;
		for (int i=0; i<params.length; i++) {
			if (!isFixed[i]) {
				if (isBoundedLo[i] && isBoundedHi[i]) {
					// High and low bounds have both been set, encode periodic boundary conditions
					double center = (boundsHi[i] + boundsLo[i])/2;
					double radius = (boundsHi[i] - boundsLo[i])/2;
					eParams[index] = Math.asin((params[i]-center)/radius);
				}
				else if (isBoundedLo[i] && !isBoundedHi[i]) {
					// Low bounds only has been set, encode square root of dist. above low boundary
					eParams[index] = Math.sqrt(params[i]-boundsLo[i]);
				}
				else if (!isBoundedLo[i] && isBoundedHi[i]) {
					// High bounds only has been set, encode square root of dist. below high boundary
					eParams[index] = Math.sqrt(boundsHi[i]-params[i]);
				}
				else {
					// No bounds set, pass the parameter
					eParams[index] = params[i];
				}
				index++;
			}
		}
		return eParams;
	}
	
	public double[] decodeParamValues(double[] eParams, boolean[] isFixed, double[] boundsLo, double[] boundsHi) {
//	public double[] decodeParamValues(double[] eParams, boolean[] isFixed) {
//		int index = 0;
		int np = params.length;
		double[] dParams = new double [np];
		for (int i=0; i<np; i++) {
			if (!isFixed[i]) {
				if (isBoundedLo[i] && isBoundedHi[i]) {
					// High and low bounds have both been set, decode periodic boundary conditions
					double center = (boundsHi[i] + boundsLo[i])/2;
					double radius = (boundsHi[i] - boundsLo[i])/2;
					dParams[i] = Math.sin(eParams[i])*radius + center;
				}
				else if (isBoundedLo[i] && !isBoundedHi[i]) {
					// Low bounds only has been set, decode square root of dist. above low boundary
					dParams[i] = eParams[i]*eParams[i] + boundsLo[i];
				}
				else if (!isBoundedLo[i] && isBoundedHi[i]) {
					// High bounds only has been set, decode square root of dist. below high boundary
					dParams[i] = boundsHi[i] - eParams[i]*eParams[i];
				}
				else {
					// No bounds set, pass the parameter
					dParams[i] = eParams[i];
				}
//				index++;
			}
			else dParams[i] = eParams[i];
		}
//		for (int i=0; i<dParams.length; i++) {
//			System.out.println("SModel.decodeParamValues: i="+i+"; dP="+dParams[i]);
//		}

		return dParams;
	}
	
//	public double[] makePixels (int length) {
//		return makePixels(params,length);
//	}
	
	abstract public int getLengthIndex();
	abstract public Point2D.Double[] makePixels (double[] p, int length);
	public Point2D.Double[] makePixels(double[] p) {
		int length = (int)Math.max(SModel.MIN_LENGTH,2*p[getLengthIndex()]+7);
		return makePixels(p,length);
	}
	abstract public Point2D.Double[] makeDisplay (double[] p);
	public Point2D.Double[] makeDisplay() {
		return makeDisplay(getParams());
	}
	abstract public Object clone();
}

